import geojson as gj
import pytest

from stelar.client.proxy.decl import ProxyState
from stelar.client.spatial import GeoJSON


def test_geojson_field():
    f = GeoJSON(nullable=True)

    f.validate(None)
    f.validate(gj.utils.generate_random("Point"))
    f.validate(gj.utils.generate_random("Polygon"))
    f.validate(gj.utils.generate_random("LineString"))


@pytest.mark.parametrize(
    "value",
    [
        "foo",
        "Point",
        [3.2, 1.4],
        1.32,
        {"type": "Point"},
        {"type": "Point", "coordinates": [13]},
        {"type": "Point", "coordinates": []},
        {"type": "Polygon", "coordinates": [[1, 2], [3, 4], [5, 6], [1, 2]]},
        {"type": "Polygon", "coordinates": [[[1, 2], [3, 4], [5, 6], [2, 2]]]},
    ],
)
def test_geojson_field_invalid(value):
    f = GeoJSON(nullable=True)

    with pytest.raises(ValueError):
        f.validate(value)


def test_geojson_to_proxy():
    f = GeoJSON(nullable=True)

    assert f.convert_to_proxy(None) is None
    assert f.convert_to_proxy(gj.utils.generate_random("Point")).type == "Point"
    assert f.convert_to_proxy(gj.utils.generate_random("Polygon")).type == "Polygon"
    assert (
        f.convert_to_proxy(gj.utils.generate_random("LineString")).type == "LineString"
    )


def test_dataset_create_spatial(testcli):
    c = testcli

    try:
        c.DELETE("v2/datasets/test_dataset?purge=true")
    except Exception:
        pass

    d = c.datasets.create(name="test_dataset", title="A test")
    assert d.title == "A test"
    assert d.spatial is None
    assert c.api.dataset_show("test_dataset")["spatial"] is None

    with pytest.raises(ValueError):
        d.spatial = "foo"
    assert d.proxy_state is ProxyState.CLEAN
    with pytest.raises(ValueError):
        d.spatial = gj.Point()
    assert d.proxy_state is ProxyState.CLEAN

    poly = gj.utils.generate_random("Polygon")
    d.spatial = poly
    assert d.spatial == poly
    assert c.api.dataset_show("test_dataset")["spatial"] == poly

    d.delete(purge=True)



from stelar.client import Client, ProxyState


def test_standard_orgganization(testcli):
    org = testcli.organizations["stelar-klms"]
    assert org.name == "stelar-klms"


def test_create_org(testcli):
    for i in range(5):
        try:
            # Use the API to delete
            testcli.DC.organization_purge(id=f"test_org_{i}")
        except:
            pass

    norg = len(testcli.organizations[::])

    orgs = []
    for i in range(5):
        print(i)
        o = testcli.organizations.create(name=f"test_org_{i}")
        orgs.append(o)

    for i in range(5):
        assert testcli.organizations[f"test_org_{i}"] is orgs[i]

    for o in orgs:
        o.delete(purge=True)

    assert len(testcli.organizations[::]) == norg


def test_generic_proxy_create(testcontext):
    c = Client(testcontext)
    org = c.organizations["stelar-klms"]

    if "test_dataset" in c.datasets:
        c.datasets["test_dataset"].delete(purge=True)

    d = c.datasets.create(
        name="test_dataset",
        title="Test dataset",
        notes="A very simple dataset to test",
        maintainer="Alonso Church",
        url="https://stelar.tuc.gr/data",
    )

    assert d.name == "test_dataset"
    assert d.title == "Test dataset"
    assert d.notes == "A very simple dataset to test"
    assert d.maintainer == "Alonso Church"
    assert d.url == "https://stelar.tuc.gr/data"

    assert d.organization is c.organizations["stelar-klms"]
    assert len(d.resources) == 0

    d.delete(purge=True)
    assert d.proxy_state is ProxyState.ERROR



from datetime import datetime
from uuid import UUID, uuid4

import pytest
from proxy_utils import ProxyTestObj, TPCatalog

from stelar.client import deferred_sync
from stelar.client.proxy import (
    BoolField,
    DateField,
    Id,
    IntField,
    InvalidationError,
    Property,
    Proxy,
    ProxyState,
    Schema,
    StrField,
    UUIDField,
)


##################################################
#  An 'abstract' subclass of Proxy which does
#  not implement an entity.
##################################################
class TestProxy(Proxy, entity=False):
    __test__ = False

    data = {}

    def proxy_sync(self, entity=None):
        if self.proxy_changed is not None:
            entity = self.data = self.proxy_to_entity()
            self.patch = self.proxy_to_entity(self.proxy_changed)
        if entity is None:
            entity = self.data
        self.proxy_from_entity(entity)
        self.proxy_changed = None


def test_abstract():
    class Foo(Proxy, entity=False):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    class Bar(Foo, entity=False):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    class Baz(Bar):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        a = Property()

        def proxy_sync(self):
            self.proxy_attr = {"a": None}
            self.proxy_changed = None

    assert not hasattr(Foo, "proxy_schema")
    assert not hasattr(Bar, "proxy_schema")
    assert hasattr(Baz, "proxy_schema")

    catalog = TPCatalog()
    PFoo = catalog.registry_for(Foo)
    PBar = catalog.registry_for(Bar)
    PBaz = catalog.registry_for(Baz)

    x = PFoo.fetch()
    assert not hasattr(x, "id")

    x = PBar.fetch()
    assert not hasattr(x, "id")

    x = PBaz.fetch()
    assert hasattr(x, "id")
    assert hasattr(x, "a")

    with pytest.raises(TypeError):

        class Bad(Baz):
            a = Property()


def test_non_entity_class_property_raises():
    with pytest.raises((RuntimeError, TypeError)):

        class Foo:
            a = Property()

    with pytest.raises((RuntimeError, TypeError)):

        class Foo2:
            a = Id()


def test_proxy_schema_registration():
    class Foo(Proxy):
        pass

    assert hasattr(Foo, "proxy_schema")
    assert Schema.for_entity("Foo") is Foo.proxy_schema


def test_empty_proxy_obj_schema():
    class Foo(Proxy):
        pass

    PFoo = TPCatalog().registry_for(Foo)
    eid = uuid4()
    x = PFoo.fetch(eid)

    #
    assert x.proxy_attr is None
    assert len(x.proxy_schema.properties) == 0
    assert x.proxy_schema.id.name == "id"
    assert x.id == eid


def test_empty_proxy_obj_given_key():
    class Foo(Proxy):
        myid = Id()

    eid = uuid4()
    x = TPCatalog().registry_for(Foo).fetch(eid)

    #
    assert x.proxy_attr is None
    assert len(x.proxy_schema.properties) == 0
    assert x.proxy_schema.id.name == "myid"
    assert x.proxy_schema.id.entity_name == "myid"
    assert x.myid == eid


def test_empty_proxy_obj_given_key_with_args():
    class Foo(Proxy):
        myid = Id(entity_name="entity_id", doc="The ID doc")

    eid = uuid4()
    x = TPCatalog().registry_for(Foo).fetch(eid)

    #
    assert x.proxy_attr is None
    assert len(x.proxy_schema.properties) == 0
    assert x.proxy_schema.id.name == "myid"
    assert x.myid == eid
    assert x.proxy_schema.id.entity_name == "entity_id"
    assert "The ID doc" in x.proxy_schema.id.__doc__


def test_property():
    class Foo(Proxy):
        aprop = Property()

    eid = uuid4()
    x = TPCatalog().registry_for(Foo).fetch(eid)

    # Just checking...
    assert x.proxy_attr is None
    assert len(x.proxy_schema.properties) == 1
    assert x.proxy_schema.id.name == "id"
    assert x.id == eid

    p = x.proxy_schema.properties["aprop"]
    assert p.name == "aprop"
    assert p.updatable is False
    assert p.optional is False
    assert p.entity_name == "aprop"


def test_proxy_obj():
    class Foo(Proxy):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.data = {
                "a": 10,
                "b": "Hello",
                "c": None,
            }

        a = Property(updatable=True)
        b = Property(updatable=True)
        c = Property(updatable=True)

        def proxy_sync(self, entity=None):
            if self.proxy_changed is not None:
                entity = self.data = self.proxy_to_entity()
                self.proxy_changed = None
            if entity is None:
                entity = self.data
            self.proxy_from_entity(entity)

    eid = uuid4()
    x = TPCatalog().registry_for(Foo).fetch(eid)

    assert x.proxy_attr is None
    assert x.proxy_state is ProxyState.EMPTY
    assert len(x.proxy_schema.properties) == 3
    assert x.proxy_schema.id.name == "id"
    assert x.id == eid
    assert x.proxy_attr is None
    assert x.proxy_changed is None

    assert x.a == 10
    assert x.b == "Hello"
    assert x.c is None
    assert x.proxy_changed is None
    assert x.proxy_state is ProxyState.CLEAN

    x.proxy_autosync = False
    x.a = 20
    x.b = "Hi there"
    assert x.proxy_changed is not None
    assert x.proxy_state is ProxyState.DIRTY
    x.proxy_sync()
    assert x.proxy_state is ProxyState.CLEAN

    assert x.a == 20
    assert x.b == "Hi there"
    assert x.proxy_changed is None
    assert x.proxy_state is ProxyState.CLEAN

    x.a = 30
    assert x.proxy_state is ProxyState.DIRTY
    assert x.a == 30
    with pytest.raises(InvalidationError):
        x.proxy_invalidate()
    x.proxy_reset()
    assert x.proxy_state is ProxyState.CLEAN
    assert x.a == 20
    assert x.b == "Hi there"

    x.a = 30
    assert x.proxy_state is ProxyState.DIRTY
    assert x.a == 30
    x.proxy_invalidate(force=True)
    assert x.proxy_state is ProxyState.EMPTY
    assert x.a == 20
    assert x.proxy_state is ProxyState.CLEAN


def test_attr_read_only():
    class Foo(TestProxy):
        a = Property(updatable=False)
        data = {"a": 20}

    eid = uuid4()
    x = TPCatalog().registry_for(Foo).fetch(eid)

    assert x.a == 20
    with pytest.raises(expected_exception=AttributeError):
        x.a = 10

    with pytest.raises(expected_exception=AttributeError):
        del x.a

    y = TPCatalog().registry_for(Foo).fetch()
    with pytest.raises(expected_exception=AttributeError):
        y.a = 10

    with pytest.raises(expected_exception=AttributeError):
        del y.a

    assert x.a == 20
    assert y.a == 20


def test_prop_optional():
    class Foo(TestProxy):
        c = Property(updatable=True)
        a = Property(optional=True, updatable=True)
        b = Property(optional=True, updatable=False)

        data = {}

    eid = uuid4()
    x = TPCatalog().registry_for(Foo).fetch(eid)
    with pytest.raises(ValueError):
        x.c

    x.data = {"c": 10}
    x.proxy_invalidate()
    assert x.c == 10
    assert not hasattr(x, "a")
    assert not hasattr(x, "b")

    x.proxy_autosync = False
    x.a = 10
    assert x.a == 10
    assert hasattr(x, "a")
    assert not hasattr(x, "b")

    assert "a" not in x.data
    x.proxy_sync()
    assert x.data["a"] == 10

    with pytest.raises(AttributeError):
        x.b = 20  # this fails because b is not updatable

    x.data["b"] = 20
    x.proxy_sync()
    assert x.b == 20
    del x.b
    assert not hasattr(x, "b")
    x.proxy_sync()
    assert "b" not in x.data


def test_property_validator():
    class Foo(TestProxy):
        a = Property(validator=IntField, updatable=True)
        b = Property(validator=StrField(), updatable=False)
        data = {"a": -1, "b": "one"}

    assert Foo.proxy_schema.properties["a"].validator.validate(10) == 10
    assert Foo.proxy_schema.properties["b"].validator.validate("10") == "10"

    x = TPCatalog().registry_for(Foo).fetch()
    assert x.a == -1
    x.a = 10
    assert x.a == 10


def test_missing_values():
    class Foo(TestProxy):
        a = Property(updatable=True)
        b = Property(updatable=False)
        data = {}

    x = TPCatalog().registry_for(Foo).fetch()
    with pytest.raises(ValueError):
        x.a

    x.data = {"a": 10, "b": 20}
    x.proxy_invalidate()
    assert x.a == 10
    assert x.b == 20


def test_set_changed_once():
    class Foo(TestProxy):
        a = Property(updatable=True)
        data = {"a": 1}

    x = TPCatalog().registry_for(Foo).fetch()
    x.proxy_autosync = False
    assert x.a == 1
    x.a = 2
    assert x.a == 2
    x.a = 3
    assert x.a == 3

    assert x.proxy_attr["a"] == 3
    assert x.proxy_changed["a"] == 1

    x.proxy_reset()
    assert x.a == 1


def test_multiple_keys_raise():
    with pytest.raises(TypeError):

        class Foo(TestProxy):
            id1 = Id()
            id2 = Id()


def test_id_property_nonkey_raise():
    with pytest.raises(TypeError):

        class Foo(TestProxy):
            id = Property()

        # This should raise
        TPCatalog().registry_for(Foo).fetch()


def test_Proxy_init():
    class Foo(TestProxy):
        a = Property()
        data = {"a": 1}

    eid = uuid4()
    eid2 = uuid4()
    tpc = TPCatalog().registry_for(Foo)

    with pytest.raises(ValueError):
        x = Foo(tpc)

    with pytest.raises(ValueError):
        x = Foo(tpc, eid=eid, entity={"id": str(eid2), "a": 10})

    with pytest.raises(ValueError):
        x = Foo(tpc, entity={"id2": str(eid2), "a": 10})

    x = Foo(tpc, eid, entity={"id": str(eid), "a": 20})
    assert x.id == eid

    x = Foo(tpc, eid, entity={"a": 30})
    assert x.id == eid
    assert x.a == 1

    x = Foo(tpc, eid)
    assert x.id == eid

    x = Foo(tpc, str(eid))
    assert x.id == eid
    assert x.a == 1

    x = Foo(tpc, entity={"id": eid, "a": 20})
    assert x.a == 1


def test_setting_on_new_obj_syncs():
    class Foo(TestProxy):
        a = Property(updatable=True)
        b = Property(updatable=True)
        data = {"a": 10, "b": 20}

    x = TPCatalog().registry_for(Foo).fetch()
    assert x.proxy_attr is None
    x.a = 1
    assert x.proxy_attr["a"] == 1
    assert x.proxy_attr["b"] == 20


def test_cannot_set_to_ellipsis():
    class Foo(TestProxy):
        a = Property(updatable=True)
        b = Property(updatable=True)
        data = {"a": 10, "b": 20}

    x = TPCatalog().registry_for(Foo).fetch()
    assert x.proxy_attr is None
    x.proxy_sync()

    with pytest.raises(ValueError):
        x.a = ...


def test_double_deletion():
    class Foo(TestProxy):
        a = Property(updatable=True)
        b = Property(updatable=True, optional=True)
        data = {"a": 10, "b": 20}

    x = TPCatalog().registry_for(Foo).fetch()
    assert x.proxy_attr is None
    x.proxy_sync()

    del x.b
    with pytest.raises(AttributeError):
        del x.b

    x.proxy_sync()
    with pytest.raises(AttributeError):
        del x.b


def test_key_unchangable():
    class Foo(TestProxy):
        a = Property(updatable=True)
        b = Property(updatable=True, optional=True)
        data = {"a": 10, "b": 20}

    eid = uuid4()
    x = TPCatalog().registry_for(Foo).fetch(eid)
    assert x.id == eid
    with pytest.raises(AttributeError):
        x.id = uuid4()

    with pytest.raises(AttributeError):
        del x.id


def test_autosync():
    class Foo(TestProxy):
        a = Property(updatable=True)
        b = Property(updatable=True, optional=True)
        data = {"a": 10, "b": 20}

    eid = uuid4()
    x = TPCatalog().registry_for(Foo).fetch(eid)

    assert x.proxy_autosync

    x.a = 4
    assert x.a == 4
    assert x.data["a"] == 4

    x.update(a=0, b="haha")
    assert x.a == 0
    assert x.b == "haha"
    assert x.data == {"a": 0, "b": "haha"}

    x.update(a=100, b=...)
    assert x.a == 100
    assert not hasattr(x, "b")
    assert x.data == {"a": 100}


def test_autosync2():
    class Foo(TestProxy):
        a = Property(validator=IntField, updatable=True, optional=True)
        b = Property(validator=IntField, updatable=True, optional=True)
        data = {"a": 10, "b": 20}

    eid = uuid4()
    x = TPCatalog().registry_for(Foo).fetch(eid)

    assert x.proxy_autosync
    with pytest.raises(ValueError):
        x.a = "aa"
    assert x.a == 10
    assert x.b == 20

    with pytest.raises(ValueError):
        x.update(a=0, b="aa")
    assert x.a == 10
    assert x.b == 20

    with pytest.raises(ValueError):
        x.update(a="aa", b=0)
    assert x.a == 10
    assert x.b == 20

    with pytest.raises(ValueError):
        x.update(a=20, b="aa")
    assert x.a == 10
    assert x.b == 20


def test_proxy_new():
    class Foo(ProxyTestObj):
        a = Property(validator=IntField, updatable=True, optional=True)
        b = Property(validator=StrField(default="world"), updatable=True, optional=True)
        c = Property(validator=BoolField(default=True), updatable=True, optional=True)
        d = Property(validator=DateField, updatable=True, optional=True)
        e = Property(
            validator=UUIDField(default=uuid4()), updatable=True, optional=True
        )

        data = {}

    uu = uuid4()
    Foo.data[uu] = Foo.new_entity(
        a=10, b="hello", c=False, d=datetime(2005, 9, 18, 23, 55), e=uu
    )

    assert Foo.data[uu]["a"] == 10
    assert Foo.data[uu]["b"] == "hello"
    assert Foo.data[uu]["c"] is False
    assert Foo.data[uu]["d"] == datetime(2005, 9, 18, 23, 55).isoformat()
    assert Foo.data[uu]["e"] == str(uu)

    c = TPCatalog()
    x = c.registry_for(Foo).fetch(uu)

    assert x.a == 10
    assert x.b == "hello"
    assert x.c is False
    assert x.d == datetime(2005, 9, 18, 23, 55)
    assert x.e == uu

    dnow = datetime.now()
    y = Foo.new(c, a=25, b="vsam", c=True, d=dnow, e=uu)

    assert y.id != UUID(int=0)
    assert y.a == 25
    assert y.b == "vsam"
    assert y.c is True
    assert y.d == dnow
    assert y.e == uu

    y.proxy_invalidate()
    assert y.proxy_state is ProxyState.EMPTY

    # Reloads
    assert y.a == 25
    assert y.b == "vsam"
    assert y.c is True
    assert y.d == dnow
    assert y.e == uu

    z = Foo.new(c, a=25, autosync=False)
    assert z.id == UUID(int=0)
    assert z.a == 25
    assert z.proxy_state is ProxyState.DIRTY

    with pytest.raises(NotImplementedError):
        z.proxy_sync()

    assert z.proxy_autosync

    with deferred_sync(z):
        assert z.id == UUID(int=0)
        z.d = dnow
        assert z.id == UUID(int=0)

    assert z.id != UUID(int=0)
    assert z.proxy_state is ProxyState.CLEAN



from .generic import GenericProxy
from .proxy import ExtrasProperty, ExtrasProxy, NameId, Property, StateField, StrField


class NamedProxy(GenericProxy, ExtrasProxy, entity=False):
    """Proxies for entities with a name which is unique.

    Named proxies include those implemented in CKAN as packages or groups.
    Other entities may have a name atrribute, but it is not unique.
    """

    name = NameId()
    state = Property(validator=StateField)
    type = Property(validator=StrField)

    # Named entities have the same 'extras' field (unlike resources)
    extras = ExtrasProperty()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from .proxy import Proxy

TagSpecList = tuple[str, ...]
TagDictList = list[dict[str, Any]]
ProxyClass = TypeVar("ProxyClass", bound="Proxy")
Entity = dict[str, Any]



"""
    Implement an accessor for CKAN API that bypasses the STELAR API.
    This is useful for debugging and testing.
"""
from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING
from urllib.parse import urljoin

import requests

from .api_call import api_context
from .proxy import EntityNotFound, Proxy, ProxyOperationError

if TYPE_CHECKING:
    from .client import Client


class CKAN:
    """A simple-minded back door to CKAN. It assumes that CKAN
       is published by the ingress on path '/dc'.

    Use as follows:
    First, ensure that there is a context in ~/.stelar containing the field
    'ckan_apitoken'. For example:
    [apitest]
    base_url=https://stelar.foo.com
    username=joe
    password=joesecret
    ckan_apitoken: 123415343416151235jgf1gh3412jk1ljh5g1l5g....

    Assuming the context is called 'apitest' (as above), do

    dc = CKAN('apitesst')
    dc.site_read()   # This is like 'ping'
    dc.help_show(name='site_read')  # shows CKAN help
    dc.package_list()
    dc.package_show(id='shakespeare_novels')

    dc.package_create(name='just_a_test', title='Just a test', owner_org='stelar-klms')
    dc.package_delete(id='just_a_test')

    ... etc. Use
    """

    def __init__(self, context="apitest", client=None):
        from .client import Client

        if client is None:
            self.client = Client(context)
        else:
            self.client = client
        self.ckanapi = urljoin(self.client._base_url, "/dc/api/3/action/")
        self.headers = {
            "Authorization": self.client._ckan_apitoken,
        }
        self.docs = {}
        self.bad_names = set()
        self.status = self.check()

    def __bool__(self) -> bool:
        return self.status

    def check(self) -> bool:
        try:
            url = urljoin(self.ckanapi, "site_read")
            resp = requests.get(url, headers=self.headers)
            result = resp.json()["result"]
            success = resp.json()["success"]
            return success and result
        except Exception:
            return False

    def __getattr__(self, name):
        if name in self.bad_names:
            raise AttributeError(name)

        # As a check that the API endpoint exists, try to fetch the
        # documentation for it. Any error in that will be translated
        # as a non-existent attribute
        try:
            doc = self.docs.get(name)
            if doc is None:
                urlh = urljoin(self.ckanapi, "help_show")
                resp = requests.post(urlh, json={"name": name}, headers=self.headers)
                doc = resp.json()["result"]
                self.docs[name] = doc
        except Exception as e:
            self.bad_names.add(name)
            raise AttributeError(name) from e

        url = urljoin(self.ckanapi, name)

        def ckan_call(json_obj={}, **kwargs):
            json = json_obj | kwargs
            resp = requests.post(url, json=json, headers=self.headers)
            return resp.json()

        ckan_call.__name__ = name
        ckan_call.__doc__ = doc
        setattr(self, name, ckan_call)

        return ckan_call

    def __repr__(self):
        return f"<CKAN {self.ckanapi} {'active' if self.status else 'bad'}>"


class api_call_DC(api_context):
    """Class that exposes the CKAN API for the Data Catalog.

    `api_call(proxy).foo(...)`
    returns the 'result' of the CKAN API response on success,
    and raises a ProxyOperationError on failure.

    `api_call(client).foo(...)`
    does the same.
    """

    def __init__(self, arg: Proxy | Client):
        super().__init__(arg)
        self.ckan = self.client.DC

    def __getattr__(self, name):
        func = getattr(self.ckan, name)

        @wraps(func)
        def wrapped_call(*args, **kwargs):
            response = func(*args, **kwargs)
            if not response["success"]:
                err = response["error"]
                if err["__type"] == "Not Found Error":
                    raise EntityNotFound(self.proxy_type, self.proxy_id, name)
                else:
                    # Generic
                    raise ProxyOperationError(
                        self.proxy_type, self.proxy_id, name, response["error"]
                    )
            return response["result"]

        return wrapped_call

    def get_call(self, proxy_type, op):
        _map_to_ckan = {
            "Dataset": "package",
            "Resource": "resource",
            "Organization": "organization",
            "Group": "group",
            "Vocabulary": "vocabulary",
            "Tag": "tag",
            "User": "user",
        }
        ckan_type = _map_to_ckan[proxy_type.__name__]
        if ckan_type == "package" and op == "purge":
            call_name = "dataset_purge"
        else:
            call_name = f"{ckan_type}_{op}"
        return getattr(self, call_name)



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Type, TypeVar
from uuid import UUID
from weakref import WeakValueDictionary

from .decl import ProxyState
from .exceptions import ConflictError
from .proxy import Proxy
from .schema import Schema

if TYPE_CHECKING:
    pass
    # from ..client import Client
    # from .property import RefList
    # This seemed to create some kind of problem...
    # from .typing import ProxyClass


ProxyClass = TypeVar("ProxyClass", bound=Proxy)


class Registry(Generic[ProxyClass]):
    """A registry is a factory and collection of proxies.

    The registry maintains a weak reference to its elements, so that
    a proxy which is not referenced any more, is deleted from the registry.

    Registries are crucial as they are used to maintain the following important
    invariant:

        'For each entity, there exists no more than one proxy (under the same client)
        at any time'.

    This invariant guarantees that, any updates to an entity (via a client)
    can only happen on a uniquely defined object. This solution saves on memory
    and avoids complex state synchronization problems.

    Registries are typed: only proxies of the same type belong to a registry.

    A collection of registries of different types form a 'catalog' (of type RegistryCatalog).
    """

    def __init__(self, catalog: RegistryCatalog, proxy_type):
        self.catalog = catalog
        self.registry = WeakValueDictionary()
        self.proxy_type = proxy_type
        if self.catalog is not None:
            self.catalog.add_registry_for(proxy_type, self)

    def fetch_proxy(self, eid: UUID) -> ProxyClass:
        """Return a proxy for the provided entity id.

        If a proxy needs to be created, it will be initialized in the
        EMPTY state by the provided entity object.

        Args:
         eid: The provided entity ID.
        Returns:
         a proxy initialized with the provided entity.
        """
        if not isinstance(eid, UUID):
            raise ValueError(f"Expected UUID, got {eid} of type {type(eid)}")
        if eid == UUID(int=0):
            raise ValueError("The null UUID(int=0) is not legal")
        proxy = self.registry.get(eid, None)
        if proxy is None:
            proxy = self.proxy_type(registry=self, eid=eid)
            assert proxy.proxy_id == eid
            self.registry[proxy.proxy_id] = proxy
        return proxy

    def fetch_proxy_for_entity(self, entity) -> ProxyClass:
        """Return a proxy for the entity object provided.

        The returned proxy will be created if needed, and
        it will be initialized in the CLEAN state by the
        provided entity object.

        Arguments:
            entity: The provided entity object.
        Returns:
            a proxy initialized with the provided entity.
        """
        eid = UUID(self.proxy_type.proxy_schema.get_id(entity))
        proxy: Proxy = self.registry.get(eid, None)
        if proxy is None:
            proxy = self.proxy_type(registry=self, entity=entity)
            assert eid == proxy.proxy_id
            assert proxy.proxy_id not in self.registry
            self.registry[proxy.proxy_id] = proxy
            proxy.proxy_sync(entity)
        else:
            if proxy.proxy_state in (ProxyState.EMPTY, ProxyState.CLEAN):
                proxy.proxy_sync(entity)
            else:
                raise ConflictError(
                    proxy,
                    entity,
                    f"Proxy fetched with new entity on state {proxy.proxy_state}",
                )
        return proxy

    def register_proxy_for_entity(self, proxy: Proxy, entity: dict):
        """Register an existing proxy ebject for the given entity.

        For success,
        1. the proxy id must be UUID(int=0)
        2. the entity must contain an id
        3. the id must not exist in the registry already.

        On success, the entity id is copied into the proxy and the
        proxy is registered.

        On failure, a conflict error is raised.

        Arguments:
            proxy: The proxy object to be registered.
            entity: The entity object to be registered.
        Returns:
            None
        Raises:
            ConflictError: If the proxy cannot be registered (is deleted or has an ID)

        """
        if proxy.proxy_id is None:
            raise ConflictError(proxy, entity, "Cannot register deleted proxy")
        if proxy.proxy_id != UUID(int=0):
            raise ConflictError(
                proxy, entity, f"Cannot register proxy with ID = {proxy.proxy_id}"
            )
        eid = UUID(self.proxy_type.proxy_schema.get_id(entity))
        if eid in self.registry:
            raise ConflictError(
                proxy, entity, f"Proxy for entity {eid} is already registered"
            )

        # Success:
        proxy.proxy_id = eid
        self.registry[eid] = proxy

    def purge_proxy(self, proxy):
        if proxy.proxy_id is None:
            # Not an error to delete something in proxy state
            return
        self.registry.pop(proxy.proxy_id, None)
        proxy.proxy_id = None

    def invalidate(self):
        """Invalidate clean entries in the registry.

        This method invalidates all CLEAN proxies in the registry,
        setting them to the EMPTY state. This will cause all subsequent
        accesses of these proxies to retrieve fresh data from the KLMS.

        The method refrains from touching dirty proxies, in order to avoid
        losing any unsaved changes.
        """
        for proxy in self.registry.values():
            if proxy.proxy_state is ProxyState.CLEAN:
                proxy.proxy_invalidate()


class RegistryCatalog:
    """Class that implements a catalog of registries, for different entity types.

    The purpose of the catalog is to provide all context information needed for
    data manipulation and transformation between proxy space and the STELAR API.

    Indeed, the RegistryCatalog class is intended to become one of the base classes
    of Client. However, it is a separate class, for better design and testability.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.registry_catalog = dict[type, Registry]()
        self.name_catalog = dict[str, Registry]()

        from .tag import VocabularyIndex

        self.vocabulary_index = VocabularyIndex(self)

    def add_registry_for(self, cls: Type[ProxyClass], registry: Registry[ProxyClass]):
        """This method adds a registry to the catalog. It is often called from inside
        the constructor of a registry.
        """
        if cls in self.registry_catalog:
            if self.registry_catalog[cls] is not registry:
                raise ValueError(
                    f"Cannot add multiple registries for entity {cls.__name__}"
                )
        else:
            self.registry_catalog[cls] = registry
            self.name_catalog[cls.__name__] = registry

    def registry_for(self, cls: Type[ProxyClass] | str) -> Registry[ProxyClass]:
        """Return the registry for the given type."""
        if isinstance(cls, str):
            return self.name_catalog[cls]
        else:
            return self.registry_catalog[cls]

    def registry_for_type(self, type_name: str) -> Registry[ProxyClass]:
        """Return the registry for the given type name.

        Entities with type name include packages, groups/organizations, resource parents.

        Args:
            type_name: The name of the type for which to return the registry.
        Returns:
            The registry for the given type name.
        """
        proxy_type_name = type_name.capitalize()
        proxy_type = Schema.for_entity(proxy_type_name).cls
        return self.registry_for(proxy_type)

    def registry_stats(self):
        """Return a Series with the number of proxies held for each entity type."""
        import pandas as pd

        return pd.Series(
            {
                r.proxy_type.__name__: len(r.registry)
                for r in self.registry_catalog.values()
            }
        )

    def fetch_active_vocabularies(self):
        """This method should return a list of dicts,
        {"name": "...",  "id": "..."}.

        More items are also allowed. It so happens that this is
        exactly the format returned by CKAN's  "vocabulary_list"
        API call. :-)
        """
        raise NotImplementedError

    def invalidate(self):
        """Invalidate all registries in the catalog.

        This method invalidates all CLEAN proxies in all registries,
        setting them to the EMPTY state. This will cause all subsequent
        accesses of these proxies to retrieve fresh data from the KLMS.
        """
        for registry in self.registry_catalog.values():
            registry.invalidate()



from .base import BaseAPI
from .policy import PolicyCursor


class AdminAPI(BaseAPI):
    """
    Represents a class that handles administrative API actions, including policy and user management.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.policies = PolicyCursor(self)



"""Declare utilities for package-derived types of entities.

These will be:
- datasets
- workflows
- workflow processes
- tools
"""
from __future__ import annotations

from functools import cached_property
from typing import Any, Optional, TypeVar
from uuid import UUID

from .api_call import api_call
from .generic import GenericCursor
from .named import NamedProxy
from .proxy import (
    BoolField,
    DateField,
    Property,
    ProxyVec,
    Reference,
    RefList,
    StrField,
    TaggableProxy,
    UUIDField,
    derived_property,
)
from .relationship import Rel, Relationship, RelProxy
from .resource import Resource
from .utils import client_for, tag_split
from .vocab import Tag


class PackageProxy(NamedProxy, TaggableProxy, entity=False):
    # Auto-maintained fields
    metadata_created = Property(validator=DateField)
    metadata_modified = Property(validator=DateField)

    creator = Property(validator=UUIDField, entity_name="creator_user_id")
    private = Property(
        validator=BoolField(nullable=False, default=False), updatable=True
    )

    organization = Reference(
        "Organization",
        entity_name="owner_org",
        create_default="default_organization",
        updatable=True,
        trigger_sync=True,
    )

    groups = RefList("Group", trigger_sync=False)

    # User-maintained fields
    notes = Property(validator=StrField(nullable=True), updatable=True)

    author = Property(validator=StrField(nullable=True), updatable=False)
    author_email = Property(validator=StrField(nullable=True), updatable=False)

    maintainer = Property(validator=StrField(nullable=True), updatable=True)
    maintainer_email = Property(validator=StrField(nullable=True), updatable=True)

    # Resources are NOT dataset-specific
    resources = RefList(Resource, trigger_sync=True)

    def add_resource(self, **properties):
        """Add a new resource with the given properties.

        Example:  new_rsrc = d.add_resource(name="Profile", url="s3://datasets/a.json",
            format="json", mimetype="application/json")

        Args:
            **properties: The arguments to pass. See 'Resource' for details.
        """
        return client_for(self).resources.create(dataset=self, **properties)

    @derived_property
    def relationships(self, entity) -> RelProxy:
        """Return a Relationships for this package.

        This is a set of relationships that this package has with other packages.
        """
        return RelProxy(self)

    def add_relationship(
        self, rel: Rel | str, obj: PackageProxy | UUID, comment: Optional[str] = None
    ):
        """Add a relationship with this package as subject.

        Args:
            rel: The relationship type to add.
            obj: The object to relate to, either a PackageProxy or UUID.
            comment: An optional comment for the relationship.

        Returns:
        """
        if isinstance(obj, PackageProxy):
            obj = str(obj.id)
        elif isinstance(obj, UUID):
            obj = str(obj)

        if isinstance(rel, Rel):
            rel = rel.value

        ac = api_call(self)
        reldata = ac.relationship_create(str(self.id), rel, obj, comment)
        return Relationship(ac.client, reldata)


PackageProxyType = TypeVar("PackageProxy", bound=PackageProxy)


class PackageCursor(GenericCursor[PackageProxyType]):
    """A cursor for package-based entities (datasets, tools, workflows etc).

    This cursor provides package-specific methods for searching and filtering
    the entities, based on CKAN's Solr backend.
    """

    def __init__(self, client, proxy_type):
        super().__init__(client, proxy_type)

    @cached_property
    def default_organization(self):
        """Return the default organization.

        This is a cached property, used in the initialization
        of the `organization` field at `PackageProxy` creation.
        """
        return client_for(self).organizations["stelar-klms"]

    def search(
        self,
        *,
        q: str | None = None,
        bbox: list[float] | None = None,
        fq: list[str] = [],
        fl: list[str] = None,
        sort: str = None,
        limit: int | None = None,
        offset: int | None = None,
        facet: dict[str, Any] | None = None,
    ):
        """
        Search for package-based entities (datasets, tools, workflows etc).

        This is the main function for searching the data catalog for packages.
        Many other functions are implemented on top of this one.

        Arguments:
            q: The query string.
            bbox: A list of four floats representing the bounding box of a spatial query.
            fq: A list of filter queries. These are not used to obtain the score of the results,
               but instead just to filter the results. They are quite efficient.
            fl: A list of fields to return. If None, a proper entity is returned.
                Note that some fields are special (not actually attribute fields), e.g.,
                the score field.
            sort: The fields to sort by. Sorting can be ascending or descending.
            limit: The maximum number of results to return.
            offset: The offset to start from.
            facet: A dictionary of facet field spec (facet attributes and limits).

        Returns:
            An answer which contains the following fields:
            - count: The number of results found (not the number of results returned).
            - results: A list of results.
            - facets: A dictionary of facets.

        """
        search = self.client.api.get_call(self.proxy_type, "search")
        args = dict(
            q=q,
            bbox=bbox,
            fq=fq,
            fl=fl,
            sort=sort,
            limit=limit,
            offset=offset,
            facet=facet,
        )
        query = {k: v for k, v in args.items() if v is not None}
        return search(query)

    def with_tag(
        self, tagarg: Tag | str, *, limit: int | None = None, offset: int | None = None
    ):
        """Return a list of entities ids which have the given tag

        Arguments:
            tagarg: A Tag (proxy) or tagspec (a string)
            limit: The maximum number of results to return.
            offset: The offset to start from.

        Returns:
            A proxy list of entity ids
        """
        # Need to obtain the tag ID, in order to call tag_show
        if isinstance(tagarg, Tag):
            return self.with_tag(tagarg.tagspec, limit=limit, offset=offset)

        # Assume that tagarg is a string
        voc, tag = tag_split(tagarg)

        if voc is None:
            filter = f'tags:"{tag}"'
        else:
            filter = f'vocab_{voc}:"{tag}"'

        rids = self.search(fq=[filter], fl=["id"], limit=limit, offset=offset)
        ids = [UUID(r["id"]) for r in rids["results"]]
        return ProxyVec(self.client, self.proxy_type, ids)



"""Utilities related to the Pandas library."""
from __future__ import annotations

import pathlib
import re
from typing import TYPE_CHECKING, Optional, Type, TypeVar
from urllib.parse import urlparse
from uuid import UUID

import pandas as pd

if TYPE_CHECKING:
    from .client import Client
    from .proxy import Proxy, ProxyList, ProxyVec

    ProxyClass = TypeVar("ProxyClass", bound=Proxy)


PANDAS_FORMATS = [
    "csv",
    "excel",
    "feather",
    "json",
    "parquet",
    "pickle",
    "stata",
    "table",
    "xml",
]

# List of pairs (regex, format) to match file extensions to formats
FILEEXT_MATCH = [
    (r"\.csv$", "csv"),
    (r"\.xls$", "excel"),
    (r"\.xlsx$", "excel"),
    (r"\.feather$", "feather"),
    (r"\.json$", "json"),
    (r"\.parquet$", "parquet"),
    (r"\.pkl$", "pickle"),
    (r"\.dta$", "stata"),
    (r"\.tsv$", "table"),
    (r"\.xml$", "xml"),
]


def infer_format(path: str, hint: str = None) -> Optional[str]:
    """Infer the format of a file from its extension, or a hint."""

    try:
        hint = hint.lower()
        if hint in PANDAS_FORMATS:
            return hint
    except Exception:
        pass

    try:
        path = urlparse(path).path
        ext = pathlib.PurePath(path).suffix
        for regex, fmt in FILEEXT_MATCH:
            if re.search(regex, ext, re.IGNORECASE):
                return fmt.lower()
        else:
            return None
    except Exception:
        return None


def get_pandas_storage_options(client: Client):
    """Get the storage options for a pandas read/write operation.

    Pandas supports access to files stored in S3, but it requires the
    credentials to be passed as a dictionary. This function returns
    the dictionary with the credentials from the client.
    """
    acc = client.s3_access_data()
    client_kwargs = {"endpoint_url": acc["endpoint"]}

    sopts = {
        "key": acc["key"],
        "secret": acc["secret"],
        "token": acc["token"],
        "client_kwargs": client_kwargs,
    }
    return sopts


def read_dataframe(client: Client, path: str, format=None, **kwargs) -> pd.DataFrame:
    """Read a DataFrame from a file.

    This function reads a DataFrame from a file and returns it.

    Parameters
    ----------"
    client : Client
        The client to use to read the file.
    path : str
        The path to the file to read. This must be an "s3" URL.
    format : str, optional
        The format of the file to read. If not specified, the format will be
        inferred from the file extension.
    kwargs : dict
        Additional keyword arguments to pass to the read"
    """
    if urlparse(path).scheme != "s3":
        raise ValueError("Only s3 URLs are supported for path")

    fmt = infer_format(path, format)
    if fmt is None:
        raise ValueError(f"Cannot infer format for file {path}")

    pdreader = getattr(pd, f"read_{fmt}")
    sopts = get_pandas_storage_options(client)
    return pdreader(path, storage_options=sopts, **kwargs)


def write_dataframe(client: Client, df: pd.DataFrame, path: str, format=None, **kwargs):
    """Write a DataFrame to a file.

    This function writes a DataFrame to a file.

    Parameters
    ----------
    client : Client
        The client to use to write the file.
    df : pd.DataFrame
        The DataFrame to write.
    path : str
        The path to the file to write.
    format : str, optional
        The format of the file to write. If not specified, the format will be
        inferred from the file extension.
    kwargs : dict
        Additional keyword arguments to pass to the write
    """
    if urlparse(path).scheme != "s3":
        raise ValueError("Only s3 URLs are supported for path")

    fmt = infer_format(path, format)
    if fmt is None:
        raise ValueError(f"Cannot infer format for file {path}")

    pdwriter = getattr(df, f"to_{fmt}")
    sopts = get_pandas_storage_options(client)
    return pdwriter(path, storage_options=sopts, **kwargs)


@pd.api.extensions.register_series_accessor("stelar")
class StelarSeriesAccessor:
    def __init__(self, ds: pd.Series):
        self.ds = ds

    def to_proxy_vec(self, client: Client, proxy_type: Type[ProxyClass]) -> ProxyList:
        """Convert a series of UUIDs to a proxy vector.

        All items in the series are expected to be UUIDs. In particular,
        None, NA and missing values are not allowed.
        """
        if not self.ds.map(lambda x: isinstance(x, UUID)).all():
            raise ValueError("The series must contain only UUIDs")
        return ProxyVec(client, proxy_type, self.ds.to_list())

    def getattr(self, name, default=pd.NA):
        """Retrieve an attribute from all objects in a series.

        This is particularly useful when the series contains proxy objects.
        """
        return self.ds.map(lambda x: getattr(x, name, default))


@pd.api.extensions.register_dataframe_accessor("stelar")
class StelarDataFrameAccessor:
    def __init__(self, df: pd.DataFrame):
        self.df = df

    def write_dataset(self, client: Client, path: str, format=None, **kwargs):
        """Write a DataFrame to a file.

        This function writes a DataFrame to a file.

        Parameters
        ----------
        client : Client
            The client to use to write the file.
        path : str
            The path to the file to write.
        format : str, optional
            The format of the file to write. If not specified, the format will be
            inferred from the file extension.
        kwargs : dict
            Additional keyword arguments to pass to the write
        """
        return write_dataframe(client, self.df, path, format, **kwargs)



from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from .client import Client
from .dataset import Dataset
from .group import Group, Organization
from .process import Process
from .proxy import ProxyState, deferred_sync

# Include all proxy exceptions
from .proxy.exceptions import *
from .proxy.exceptions import __all__ as __all_exceptions
from .relationship import Rel, Relationship, Relationships
from .resource import Resource
from .task_spec import TaskSpec
from .tasks import Task
from .tool import Tool
from .user import User

# Include all utility functions
from .utils import *
from .utils import __all__ as __all_utils
from .vocab import Tag, Vocabulary
from .workflows import Workflow

__all__ = [
    *__all_exceptions,
    *__all_utils,
    "Client",
    "Dataset",
    "Resource",
    "Organization",
    "Group",
    "ProxyState",
    "Tag",
    "Vocabulary",
    "User",
    "Process",
    "Relationship",
    "Rel",
    "Relationships",
    "Task",
    "TaskSpec",
    "Workflow",
    "Tool",
    "deferred_sync",
]

del __all_exceptions
del __all_utils

import os

if os.getenv("SPHINX_BUILD"):
    # This prevents duplicate documentation by sphinx
    del __all__



"""Generation of rich (e.g., html) representations for proxies.

Code used here follows the ipython rich display protocol, which is
described in the IPython documentation.

It is also useful in the context of Jupyter notebooks, where the
representation of objects can be customized to provide a richer
experience to the user.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from IPython.display import HTML

if TYPE_CHECKING:
    from .dataset import Dataset
    from .resource import Resource


def resource_to_html(obj: Resource) -> str:
    """Return an HTML representation of a resource.

    Provide an HTML representation of the Resource instance for Jupyter display,
    with enhanced styles, watermark, and consistent formatting.
    """
    # Define the watermark image URL (Replace with your image path)
    watermark_image_url = "logo.png"

    # Build the table HTML
    prop_style = (
        "text-align: left; padding: 5px; border: 1px solid #ddd; font-weight: bold;"
    )
    val_style = "text-align: left; padding: 5px; border: 1px solid #ddd;"

    html = f"""
    <div style="position: relative; width: 100%; height: auto; display: flex; justify-content:
                                flex-start; align-items: flex-start; margin-bottom: 20px;">
        <table border="1" style="
            border-collapse: collapse;
            width: 50%;
            margin: 0;
            color: black;
            background-image: url('{watermark_image_url}');
            background-size: 20%;
            background-position: center;
            background-repeat: no-repeat;">
            <thead>
                <tr>
                    <th colspan="2" style="text-align: center; padding: 10px; font-size: 18px;
                    font-weight: bold; background-color: rgba(242, 242, 242, 0.8);">
                        Resource Summary
                    </th>
                </tr>
                <tr style="background-color: rgba(242, 242, 242, 0.8);">
                    <th style="width: 50%; text-align: center;">Attribute</th>
                    <th style="width: 50%; text-align: center;">Value</th>
                </tr>
            </thead>
            <tbody>
                <tr style="background-color: rgba(255, 255, 255, 0.8);">
                    <td style=>ID:</td>
                    <td style={val_style}>{obj.id or 'N/A'}</td>
                </tr>
                <tr style="background-color: rgba(230, 179, 255, 0.8);">
                    <td style={prop_style}>Parent Package:</td>
                    <td style={val_style}>{obj.dataset.id or 'N/A'}</td>
                </tr>
                <tr style="background-color: rgba(255, 255, 255, 0.8);">
                    <td style={prop_style}>Relation To Parent:</td>
                    <td style={val_style}>{obj.relation or 'N/A'}</td>
                </tr>
                <tr style="background-color: rgba(230, 179, 255, 0.8);">
                    <td style={prop_style}>Name:</td>
                    <td style={val_style}>{obj.name}</td>
                </tr>
                <tr style="background-color: rgba(255, 255, 255, 0.8);">
                    <td style={prop_style}>URL:</td>
                    <td style={val_style}><a href="{obj.url}" target="_blank">{obj.url}</a></td>
                </tr>
                <tr style="background-color: rgba(230, 179, 255, 0.8);">
                    <td style={prop_style}>Format:</td>
                    <td style={val_style}>{obj.format}</td>
                </tr>
                <tr style="background-color: rgba(255, 255, 255, 0.8);">
                    <td style={prop_style}>Description:</td>
                    <td style={val_style}>{obj.description or 'N/A'}</td>
                </tr>
            </tbody>
        </table>
    </div>
    """
    return HTML(html)._repr_html_()


def dataset_to_html(obj: Dataset) -> str:
    """
    Provide an HTML representation of the Dataset instance for Jupyter display,
    with enhanced styles, watermark, and consistent formatting.
    """

    prop_style = (
        "text-align: left; padding: 5px; border: 1px solid #ddd; font-weight: bold;"
    )
    val_style = "text-align: left; padding: 5px; border: 1px solid #ddd;"

    # Build the HTML for the resources table
    resources_html = ""
    if obj.resources:
        resources_html = "".join(
            f"""
                <tr style="background-color: 
                    {('rgba(255, 255, 255, 0.8)' if i % 2 == 0 else 'rgba(230, 179, 255, 0.8)')};">
                    <td style={val_style}>{resource.id or 'N/A'}</td>
                    <td style={val_style}>{resource.name}</td>
                    <td style={val_style}><a href="{resource.url}" target="_blank">{resource.url}</a></td>
                    <td style={val_style}>{resource.format}</td>
                </tr>
                """
            for i, resource in enumerate(obj.resources)
        )
    else:
        resources_html = """
                <tr style="background-color: rgba(255, 255, 255, 0.8);">
                    <td colspan='5' style="text-align: center; padding: 10px; 
                            border: 1px solid #ddd;">No Resources Associated</td>
                </tr>
            """

    # Define the watermark image URL (Replace with your image path)
    watermark_image_url = "logo.png"

    # Build the Dataset Summary table
    summary_html = f"""
        <div style="position: relative; width: 100%; height: auto; 
                    display: flex; justify-content: flex-start; 
                    align-items: flex-start; margin-bottom: 20px;">
            <table border="1" style="
                border-collapse: collapse;
                width: 50%;
                margin: 0;
                color: black;
                background-image: url('{watermark_image_url}');
                background-size: 20%;
                background-position: center;
                background-repeat: no-repeat;">
                <thead>
                    <tr>
                        <th colspan="2" style="text-align: center; 
                                padding: 10px; font-size: 18px; 
                                font-weight: bold; 
                                background-color: rgba(242, 242, 242, 0.8);">
                            Dataset Summary
                        </th>
                    </tr>
                </thead>
                <tbody>
                    <tr style="background-color: rgba(255, 255, 255, 0.8);">
                        <td style={prop_style}>ID:</td>
                        <td style={val_style}>{obj.id or 'N/A'}</td>
                    </tr>
                    <tr style="background-color: rgba(230, 179, 255, 0.8);">
                        <td style={prop_style}>Title:</td>
                        <td style={val_style}>{obj.title}</td>
                    </tr>
                    <tr style="background-color: rgba(255, 255, 255, 0.8);">
                        <td style={prop_style}>Notes:</td>
                        <td style={val_style}>{obj.notes}</td>
                    </tr>
                    <tr style="background-color: rgba(230, 179, 255, 0.8);">
                        <td style={prop_style}>Tags:</td>
                        <td style={val_style}>{', '.join(['obj.tags'])}</td>
                    </tr>
                    <tr style="background-color: rgba(255, 255, 255, 0.8);">
                        <td style={prop_style}>Modified Date:</td>
                        <td style={val_style}>{obj.metadata_modified or 'N/A'}</td>
                    </tr>
                </tbody>
            </table>
        </div>
        """

    # Build the Dataset Resources table
    resources_table_html = f"""
        <div style="
            position: relative; 
            width: 100%; height: auto; 
            display: flex; justify-content: flex-start; align-items: start;">
            <table border="1" style="
                border-collapse: collapse;
                width: 80%;
                margin: 0;
                color: black;
                background-image: url('{watermark_image_url}');
                background-size: 20%;
                background-position: center;
                background-repeat: no-repeat;">
                <thead>
                    <tr>
                        <th colspan="5" style="text-align: center; 
                                    padding: 10px; font-size: 18px; 
                                    font-weight: bold; 
                                    background-color: rgba(242, 242, 242, 0.8);">
                            Dataset Resources
                        </th>
                    </tr>
                    <tr style="background-color: rgba(242, 242, 242, 0.8);">
                        <th style={val_style}>ID</th>
                        <th style={val_style}>Relation to Parent</th>
                        <th style={val_style}>Name</th>
                        <th style={val_style}>URL</th>
                        <th style={val_style}>Format</th>
                    </tr>
                </thead>
                <tbody>
                    {resources_html}
                </tbody>
            </table>
        </div>
        """

    # Combine both tables into the final HTML
    html = f"{summary_html}<br>{resources_table_html}"
    return HTML(html)._repr_html_()



from __future__ import annotations

from .api_call import api_call
from .package import PackageCursor, PackageProxy
from .proxy import DateField, Property, Reference, RefList, StrField
from .proxy.fieldvalidation import EnumeratedField
from .task_spec import TaskSpec
from .tasks import Task
from .utils import client_for


class ExecStateField(EnumeratedField):
    VALUES = ["running", "succeeded", "failed"]


class Process(PackageProxy):
    """Proxy object for workflow processes (executions)."""

    # Currently, the STELAR API redefines the 'creator' field to return
    # the username of the creator, rather than the user ID. This is a
    # temporary workaround until the API is updated to return the user ID.
    creator = Property(validator=StrField, entity_name="creator")

    # These are all non-CKAN properties in the STELAR API
    workflow = Reference(
        "Workflow",
        updatable=True,
        nullable=True,
        entity_name="workflow",
        trigger_sync=False,
    )
    tasks = RefList("Task", trigger_sync=True)

    start_date = Property(validator=DateField, updatable=False)
    end_date = Property(validator=DateField, updatable=True)
    exec_state = Property(validator=ExecStateField, updatable=False)

    title = Property(validator=StrField, updatable=True)
    version = Property(
        validator=StrField(nullable=True, maximum_len=100), updatable=True
    )
    url = Property(validator=StrField(nullable=True), updatable=True)

    def terminate(self, new_state: str = "succeeded"):
        """Terminate the process execution.

        This call transitions the exec_state of the process to the given state.
        The only meaningful values are those validated with the ExecStateField
        (namely, 'running', 'succeeded', 'failed').

        Note, that the process must be 'running' for this call to have any effect.

        Args:
            new_state: The new state to transition to.
        """
        # Since the exec_state is not updatable, we need to use the API directly
        entity = api_call(self).process_patch(
            id=str(self.proxy_id), exec_state="succeeded"
        )
        # Sync the proxy with the entity
        self.proxy_sync(entity=entity)

    def run(self, task_spec: TaskSpec, secrets=None) -> Task:
        """Create a new Task in this process"""
        return client_for(self).tasks.create(self.id, task_spec, secrets=secrets)


class ProcessCursor(PackageCursor[Process]):
    def __init__(self, client):
        super().__init__(client, Process)



from __future__ import annotations

from .package import PackageProxy
from .proxy import Property, StrField


class Workflow(PackageProxy):
    title = Property(validator=StrField, updatable=True)
    version = Property(
        validator=StrField(nullable=True, maximum_len=100), updatable=True
    )

    repository = Property(validator=StrField(nullable=True), updatable=True)
    executor = Property(validator=StrField(nullable=True), updatable=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Type, TypeVar
from uuid import UUID

from .decl import ProxyState
from .proxy import Proxy
from .registry import Registry

if TYPE_CHECKING:
    from ..client import Client


ProxyClass = TypeVar("ProxyClass", bound=Proxy)


class ProxyCursor(Registry[ProxyClass]):
    """Proxy cursors provide access functions to the collection of all entities of
    a given type.

    The cursor is used to fetch entities in a lazy way, and to provide
    a way to iterate over the entities, check for the existence of an entity by name,
    etc.
    """

    MAX_FETCH = 1000

    def __init__(self, client: Client, proxy_type: Type[ProxyClass]):
        """Initialize the cursor with the client and the proxy type."""
        self.client = client
        super().__init__(client, proxy_type)

    def create(self, **updates) -> ProxyClass:
        """Create a new entity of this type."""
        raise NotImplementedError

    def fetch_list(self, *, limit, offset) -> list[str]:
        """Lists entities of this type."""
        raise NotImplementedError

    def fetch(self, *, limit, offset) -> Iterator[ProxyClass]:
        """Fetch entities of this type."""
        raise NotImplementedError

    def get(self, name_or_id, default=None) -> ProxyClass:
        """Get an entity by name or ID."""
        raise NotImplementedError

    def __getitem__(self, item: str | UUID | slice) -> ProxyClass | list[ProxyClass]:
        """Get an entity by name or ID, or get a list of entities.

        A cursor behaves in a limited way as a (read-only) list of entities.
        If C is a cursor, then
        - C[UUID] returns the entity with the given UUID
        - C[str] returns the entity with the given name
        - C[:] returns a list of 'all' entities (up to some limit)
        - C[5:10] returns a list of the entities from 5 to 10, in some standard order.
        """
        if isinstance(item, slice):
            offset = item.start if item.start is not None else 0
            if offset < 0:
                raise ValueError("Bad offset")
            if item.step is not None:
                limit = item.step
            elif item.stop is not None:
                limit = item.stop - offset
            else:
                limit = self.MAX_FETCH
            if limit < 0:
                raise ValueError("Bad limit")
            # CUT OFF
            if limit > self.MAX_FETCH:
                limit = self.MAX_FETCH

            return self.fetch_list(limit=limit, offset=offset)
            # return list(self.fetch(limit=limit, offset=offset))

        elif isinstance(item, (str, UUID)):
            proxy = self.get(item)
            if proxy is None:
                raise KeyError("Entity not found")
            return proxy
        else:
            raise TypeError(
                f"Cannot fetch {self.proxy_type.__name__} by {item}: string or UUID is expected"
            )

    def __contains__(self, item) -> bool:
        """Check if an entity exists.

        The item can be a string, a UUID, or a proxy object. If it is a proxy object,
        the check is done on the proxy state not being ERROR.

        If the item is a string or a UUID, the check is done by looking the entity up
        in the STELAR API. Note that, for Data Catalog entities, a string may designate
        either an ID or a name.
        """
        if isinstance(item, self.proxy_type):
            return item.proxy_state is not ProxyState.ERROR
        elif isinstance(item, (str, UUID)):
            return self.get(item) is not None
        else:
            return False



from functools import cached_property

from .api_call import api_call
from .generic import GenericCursor, GenericProxy
from .proxy import BoolField, DateField, Id, Property, StrField


class Policy(GenericProxy):
    """
    A proxy for a STELAR policy.
    """

    id = Id(entity_name="policy_uuid")

    creator = Property(
        validator=StrField(nullable=False),
        entity_name="user_id",
        doc="The user who created the policy.",
    )

    active = Property(
        validator=BoolField(nullable=False, default=True),
        updatable=False,
        optional=False,
        doc="Whether the policy is active or not.",
    )

    created_at = Property(
        validator=DateField(nullable=False), updatable=False, optional=False
    )

    policy_name = Property(
        validator=StrField(nullable=False),
        entity_name="policy_familiar_name",
        updatable=True,
        optional=False,
        doc="The name of the policy to be applied.",
    )

    @cached_property
    def spec(self):
        """
        A property that returns the policy content as a byte array.
        """
        return api_call(self).policy_spec(str(self.proxy_id))


class PolicyCursor(GenericCursor):
    """
    A cursor for iterating over policies.
    """

    def __init__(self, client):
        super().__init__(client, Policy)

    def create(self, policy_yaml: bytes) -> Policy:
        """
        Create a new policy.

        Args:
            spec (bytes): The policy specification as a byte array.
            **kwargs: Additional keyword arguments for the policy.

        Returns:
            Policy: The created policy.
        """
        ac = api_call(self)
        result = ac.policy_create(policy_yaml=policy_yaml)
        return self.fetch_proxy_for_entity(result)



import pytest
from minio.datatypes import Object

from stelar.client.mutils import s3spec_to_pair


def test_s3objspec_to_pair():
    assert s3spec_to_pair("s3://bucket/object") == ("bucket", "object")
    assert s3spec_to_pair("s3:///object") == ("", "object")
    assert s3spec_to_pair("s3:///") == ("", "")
    assert s3spec_to_pair(("bucket", "object")) == ("bucket", "object")
    assert s3spec_to_pair(Object("bucket", "object")) == ("bucket", "object")
    assert s3spec_to_pair("s3://bucket/") == ("bucket", "")
    assert s3spec_to_pair("s3://bucket/a/bct") == ("bucket", "a/bct")
    assert s3spec_to_pair("s3://bucket") == ("bucket", "")
    assert s3spec_to_pair(("bucket", "")) == ("bucket", "")
    assert s3spec_to_pair(("bucket", "object")) == ("bucket", "object")


def test_s3objspec_to_pair_invalid():
    with pytest.raises(ValueError):
        s3spec_to_pair("http://bucket/object")
    with pytest.raises(TypeError):
        s3spec_to_pair(123)
    with pytest.raises(TypeError):
        s3spec_to_pair(("bucket", "object", "extra"))
    with pytest.raises(ValueError):
        s3spec_to_pair("//bucket/object")
    with pytest.raises(ValueError):
        s3spec_to_pair("/bucket/object")
    with pytest.raises(ValueError):
        s3spec_to_pair("bucket/object")
    with pytest.raises(ValueError):
        s3spec_to_pair("/object")
    with pytest.raises(ValueError):
        s3spec_to_pair("object")
    with pytest.raises(ValueError):
        s3spec_to_pair("")


def test_open_text_file(testcli):
    path = "s3://klms-bucket/rj.txt"

    with testcli.s3fs_open(path, mode="rb") as f:
        data = f.read()
        assert type(data) == bytes

    with testcli.s3.get_object(bucket_name="klms-bucket", object_name="rj.txt") as o:
        assert o.data == data

    with testcli.s3fs_open(path, mode="r") as f:
        data = f.read()
        assert type(data) == str


def test_create_text_file(testcli):
    path = "s3://klms-bucket/hello.txt"
    data = "Hello, World!"

    with testcli.s3fs_open(path, mode="w") as f:
        print("Hello, World!", file=f)

    with testcli.s3fs_open(path, mode="r") as f:
        assert f.read() == data + "\n"



#
# Various heplper functions that do not belong in the API
#
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Type
from uuid import UUID

from .proxy import Proxy, ProxyCursor, ProxyList, Registry

if TYPE_CHECKING:
    from .client import Client

from .proxy.decl import (
    TAGSPEC_PATTERN,
    tag_join,
    tag_split,
    validate_tagname,
    validate_tagspec,
)

__all__ = [
    "client_for",
    "tag_split",
    "tag_join",
    "validate_tagspec",
    "validate_tagname",
    "TAGSPEC_PATTERN",
]


def client_for(obj: Any) -> Client:
    """Return the client for a proxy object."""
    from .client import Client

    match obj:
        case Proxy():
            return obj.proxy_registry.catalog
        case ProxyList():
            return obj.client
        case ProxyCursor():
            return obj.client
        case Registry():
            return obj.catalog
        case Client():
            return obj
        case _:
            raise RuntimeError(
                f"Cannot resolve client for type {obj.__class__.__qualname__}", obj
            )


def convert_proxy_id_to_str(
    value: Proxy | UUID | str, proxy_type: Type = Proxy, *, nullable: bool = True
) -> str | None:
    """Convert a UUID or Proxy to a string representation."""
    if nullable:
        if value is None:
            return None
    if isinstance(value, proxy_type):
        if value.proxy_id is None:
            raise ValueError("Proxy has no ID")
        return str(value.proxy_id)
    elif isinstance(value, UUID):
        return str(value)
    elif isinstance(value, str):
        return value
    else:
        raise TypeError(f"Cannot convert proxy {value} of type {type(value)} to string")



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional
from uuid import UUID

from .api_call import api_call

if TYPE_CHECKING:
    from .client import Client
    from .package import PackageProxy


class Rel(Enum):
    """Enum for relationship types."""

    PARENT_OF = "parent_of"
    CHILD_OF = "child_of"
    DEPENDS_ON = "depends_on"
    DEPENDENCY_OF = "dependency_of"
    DERIVES_FROM = "derives_from"
    HAS_DERIVATION = "has_derivation"
    LINKS_TO = "links_to"
    LINKED_FROM = "linked_from"

    def is_canonical(self) -> bool:
        """Return True if the relationship is forward, False if it is backward."""
        return self in CANONICAL

    def peer(self) -> Rel:
        """Return the peer relationship type."""
        return PEER_MAP[self]


CANONICAL = [Rel.CHILD_OF, Rel.DEPENDS_ON, Rel.DERIVES_FROM, Rel.LINKS_TO]
PEERS = [
    (Rel.PARENT_OF, Rel.CHILD_OF),
    (Rel.DEPENDENCY_OF, Rel.DEPENDS_ON),
    (Rel.HAS_DERIVATION, Rel.DERIVES_FROM),
    (Rel.LINKS_TO, Rel.LINKED_FROM),
]
PEER_MAP = {rel1: rel2 for rel1, rel2 in PEERS} | {rel2: rel1 for rel1, rel2 in PEERS}


class Relationship:
    def __init__(self, client: Client, reldata: Dict[str, Any]) -> None:
        self.client: Client = client
        self.subject_id: UUID = UUID(reldata["subject"])
        self.subject_name: str = reldata["subject_name"]
        self.subject_type: str = reldata["subject_type"]

        self.object_id: UUID = UUID(reldata["object"])
        self.object_name: str = reldata["object_name"]
        self.object_type: str = reldata["object_type"]

        self.relationship: Rel = Rel(reldata["relationship"])
        self.comment: Optional[str] = reldata.get("comment", None)

    @property
    def subject(self) -> PackageProxy:
        """Return a proxy to the subject of the relationship."""
        return self.client.registry_for_type(self.subject_type).fetch_proxy(
            self.subject_id
        )

    @property
    def object(self) -> PackageProxy:
        """Return a proxy to the object of the relationship."""
        return self.client.registry_for_type(self.object_type).fetch_proxy(
            self.object_id
        )

    def __repr__(self) -> str:
        return f"<Relationship {self.subject_name} {self.relationship.value} {self.object_name}>"

    def __str__(self) -> str:
        return f"{self.subject_name} --{self.relationship.value}--> {self.object_name}"

    def __eq__(self, other: Relationship) -> bool:
        if not isinstance(other, Relationship):
            return False
        elif self.relationship is other.relationship:
            return (
                self.subject_id == other.subject_id
                and self.object_id == other.object_id
            )
        elif self.relationship.peer() is other.relationship:
            return (
                self.subject_id == other.object_id
                and self.object_id == other.subject_id
            )
        else:
            return False

    def __hash__(self) -> int:
        """Return a hash of the relationship."""
        if self.relationship.is_canonical():
            return hash((self.subject_id, self.object_id, self.relationship))
        else:
            return hash((self.object_id, self.subject_id, self.relationship.peer()))

    def matches(
        self,
        subj: PackageProxy | UUID,
        rel: Optional[Rel] = None,
        obj: Optional[PackageProxy | UUID] = None,
    ) -> bool:
        """Check if the relationship matches the given subject, relationship type, and object.

        Matching is semantic, i.e., a PARENT_OF will match
        a CHILD_OF relationship if the subject and object are reversed.

        Args:
            subj: The subject to match against, either a PackageProxy or UUID.
            rel: The relationship type to match against, if specified.
            obj: The object to match against, either a PackageProxy or UUID.

        Returns:
            bool: True if the relationship matches the given subject and optional relationship type,
                False otherwise.
        """

        def to_uuid(item: PackageProxy | UUID) -> UUID:
            """Convert a PackageProxy or UUID to a UUID."""
            from .package import PackageProxy

            if isinstance(item, PackageProxy):
                return item.proxy_id
            elif isinstance(item, UUID):
                return item
            else:
                raise TypeError("Expected PackageProxy or UUID")

        def M(a, b) -> bool:
            """Check if two UUIDs match."""
            return b is None or a == b

        subj = to_uuid(subj) if subj is not None else None
        obj = to_uuid(obj) if obj is not None else None

        return (
            M(self.relationship, rel)
            and M(self.subject_id, subj)
            and M(self.object_id, obj)
        ) or (
            M(self.relationship.peer(), rel)
            and M(self.subject_id, obj)
            and M(self.object_id, subj)
        )

    def peer(self) -> Relationship:
        """Return the peer relationship."""
        return Relationship(
            self.client,
            {
                "subject": str(self.object_id),
                "subject_name": self.object_name,
                "subject_type": self.object_type,
                "object": str(self.subject_id),
                "object_name": self.subject_name,
                "object_type": self.subject_type,
                "relationship": self.relationship.peer().value,
                "comment": self.comment,
            },
        )

    def canonical(self) -> Relationship:
        """Return the canonical version of the relationship."""
        if self.relationship.is_canonical():
            return self
        else:
            return self.peer()

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary representation of the relationship."""
        return {
            "subject": str(self.subject_id),
            "subject_name": self.subject_name,
            "subject_type": self.subject_type,
            "object": str(self.object_id),
            "object_name": self.object_name,
            "object_type": self.object_type,
            "relationship": self.relationship.value,
            "comment": self.comment,
        }

    def from_dict(self, data: Dict[str, Any]) -> None:
        """Update the relationship from a dictionary."""
        self.subject_id = UUID(data["subject"])
        self.subject_name = data["subject_name"]
        self.subject_type = data["subject_type"]
        self.object_id = UUID(data["object"])
        self.object_name = data["object_name"]
        self.object_type = data["object_type"]
        self.relationship = Rel(data["relationship"])
        self.comment = data.get("comment", None)

    def set_comment(self, new_comment: str) -> None:
        """Set a new comment for the relationship."""
        ac = api_call(self.client)
        r = ac.relationship_update(
            str(self.subject_id),
            self.relationship.value,
            str(self.object_id),
            comment=new_comment,
        )
        self.from_dict(r)

    def delete(self) -> None:
        """Delete the relationship."""
        ac = api_call(self.client)
        ac.relationship_delete(
            str(self.subject_id), self.relationship.value, str(self.object_id)
        )


class Relationships(set):
    """A set of relationships."""

    @classmethod
    def from_api(
        cls, client: Client, reldata: Iterable[Dict[str, Any]]
    ) -> Relationships:
        """Create a Relationships from API data."""
        return cls(client, {Relationship(client, r) for r in reldata})

    def __init__(
        self, client: Client, relationships: Iterable[Relationship] = []
    ) -> None:
        super().__init__(relationships)
        self.client: Client = client

    def matching(
        self,
        subj: Optional[PackageProxy | UUID],
        rel: Optional[Rel],
        obj: Optional[PackageProxy | UUID],
    ) -> Relationships:
        """Return the relationships matching a pattern.

        This method filters the relationships in the set based on the provided
        subject, rel and object. If any of these parameters are None, they are
        ignored in the matching. Note that matching is done semantically,
        i.e., a PARENT_OF matches a CHILD_OF, reversing the subject and object.

        For this, the `matches` method of the `Relationship` class is used.

        Args:
            subj: The subject to match against, either a PackageProxy or UUID.
            rel: The relationship type to match against, if specified.
            obj: The object to match against, either a PackageProxy or UUID.

        Returns:
            Relationships: A new set of relationships that match the criteria.
        """
        return Relationships(
            self.client,
            {r for r in self if r.matches(subj, rel, obj)},
        )

    def entities(self):
        """Return a set of entities in the relationships.

        Entities are either subjects or objects of some relationship.
        """
        entset = set()
        for r in self:
            entset.add(r.subject)
            entset.add(r.object)
        return entset

    def delete(self) -> None:
        """Delete all relationships in the set."""
        for r in self:
            r.delete()
        self.clear()

    def __or__(self, other: Relationships) -> Relationships:
        """Return the union of two Relationships sets."""
        return Relationships(self.client, self | other)

    def __and__(self, other: Relationships) -> Relationships:
        """Return the intersection of two Relationships sets."""
        return Relationships(self.client, self & other)

    def __sub__(self, other: Relationships) -> Relationships:
        """Return the difference of two Relationships sets."""
        return Relationships(self.client, self - other)

    def __xor__(self, other: Relationships) -> Relationships:
        """Return the symmetric difference of two Relationships sets."""
        return Relationships(self.client, self ^ other)


class RelProxy:
    """A proxy for a set of relationships."""

    def __init__(self, subj: PackageProxy):
        self.subject = subj
        self.relationship = None

    def _to_api(self, rel: Rel | str, obj: PackageProxy | UUID) -> str:
        subj = str(self.subject.proxy_id)
        if isinstance(rel, Rel):
            rel = rel.value

        from .package import PackageProxy

        if isinstance(obj, PackageProxy):
            obj = str(obj.id)
        elif isinstance(obj, UUID):
            obj = str(obj)

        return subj, rel, obj

    def get(self, rel: Rel | str, obj: PackageProxy | UUID) -> Relationship:
        subj, rel, obj = self._to_api(rel, obj)
        ac = api_call(self.subject)
        reldata = ac.relationship_show(subj, rel, obj)
        return Relationship(ac.client, reldata)

    def __call__(self, rel: Rel | str | None = None) -> Relationships:
        """Get all relationships for the subject."""
        subj, rel, obj = self._to_api(rel, None)
        ac = api_call(self.subject)
        reldata = ac.relationships_fetch(str(self.subject.id), rel, obj)
        return Relationships.from_api(ac.client, reldata)

    def add(
        self, rel: Rel | str, obj: PackageProxy | UUID, comment: Optional[str] = None
    ) -> Relationship:
        """Add a relationship to the subject."""
        subj, rel, obj = self._to_api(rel, obj)
        ac = api_call(self.subject)
        reldata = ac.relationship_create(subj, rel, obj, comment)
        return Relationship(ac.client, reldata)

    def remove(self, rel: Rel | str, obj: PackageProxy | UUID) -> None:
        """Remove a relationship from the subject."""
        subj, rel, obj = self._to_api(rel, obj)
        ac = api_call(self.subject)
        ac.relationship_delete(subj, rel, obj)

    @property
    def subject_id(self) -> UUID:
        """Return the ID of the subject."""
        return self.subject.proxy_id

    @property
    def depends_on(self) -> Relationships:
        return self(Rel.DEPENDS_ON)

    @property
    def dependency_of(self) -> Relationships:
        return self(Rel.DEPENDENCY_OF)

    @property
    def parent_of(self) -> Relationships:
        return self(Rel.PARENT_OF)

    @property
    def child_of(self) -> Relationships:
        return self(Rel.CHILD_OF)

    @property
    def derives_from(self) -> Relationships:
        return self(Rel.DERIVES_FROM)

    @property
    def has_derivation(self) -> Relationships:
        return self(Rel.HAS_DERIVATION)

    @property
    def links_to(self) -> Relationships:
        return self(Rel.LINKS_TO)

    @property
    def linked_from(self) -> Relationships:
        return self(Rel.LINKED_FROM)



from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Optional
from uuid import UUID

from .decl import ProxyState
from .exceptions import ConversionError, ErrorState, InvalidationError

if TYPE_CHECKING:
    from pandas import Series

    from .registry import Registry, RegistryCatalog

"""
    Introduction
    ------------

    Proxy objects represent STELAR entities in Python:
    - Datasets
    - Resources
    - Workflows
    - Groups
    - Organizations
    - Processes (workflow executions)
    - Tasks
    - Tools
    - Policies
    - Users

    They can be used to
    - inspect
    - update
    - delete
    - link and relate
    - plus, custom operations

    The proxy object base class implements generic handling of properties:
    - initializing properties
    - loading propertties on demand
    - updating single properties
    - updating multiple properties

"""


# ----------------------------------------------------------
#
# Proxy is the base class for all proxy classes
#
# ----------------------------------------------------------

Entity = dict[str, Any]


class Proxy:
    """Base class for all proxy objects of the STELAR entities.

    Proxy objects are managed by Registry. The primary implementation
    of Registry is Client.

    A proxy can be in one of four states:
      - ``EMPTY``: there is no entity data in the proxy
      - ``CLEAN``: the data loaded by the last proxy_sync() operation is not changed
      - ``DIRTY``: the data loaded by the last proxy_sync() operation has been changed
      - ``ERROR``: the proxy is illegal! This state can be the result of deleting entity


    Attributes are used to hold property values:

    proxy_registry:
        The Registry instance that this proxy belongs to

    proxy_id:
        The UUID of the proxies entity

    proxy_attr:
        A dict of all loaded attributes. When None, the entity has
        not yet been loaded. The state is EMPTY

    proxy_changed:
        A dict of all changed attributes (loaded values of updated attributes)
        since last upload. When None, the entity is ``CLEAN`` (or ``EMPTY``), else the
        entity is ``DIRTY``.

    To initialize a proxy object, one must supply either an entity id or
    an entity JSON body. The proxy_id is never changed. When an entity is deleted,
    set the proxy_id to None, to mark the proxy state as ``ERROR``.

    After initialization, the state of a proxy is ``EMPTY``.

    Proxies are handlers for **entities** of the Stelar Service API. Entities are manipulated
    by a CRUD-like API. Besides creation and deletion, entities are manipulated by two additional
    I/O API operations:
    - fetch, which returns an entity data from the API
    - update, which accepts a spec of the updates to apply to an entity. This operation often returns
    the updated object after updates are applied.

    The following operations operate on proxies:

    proxy_sync(entity=None):
        Save any pending updates to make the state CLEAN. Load the proxy data
        from the Stelar Service API, to make sure the proxy has the latest. When `entity` is not None,
        use it to avoid a fetch.

    proxy_invalidate(force=False):
        Make the object EMPTY. If the proxy is DIRTY, an IvalidationError
        is raised, unless `force` is specified as True.

    proxy_reset():
        Make a DIRTY object to CLEAN, by restoring the property values of the last sync().

    proxy_sync(entity=None):
        Make an entity CLEAN.
    """

    proxy_registry: Registry
    proxy_id: Optional[UUID]
    proxy_autosync: bool
    proxy_attr: dict[str, Any] | None
    proxy_changed: dict[str, Any] | None

    def __init__(
        self, registry: Registry, eid: Optional[str | UUID] = None, entity=None
    ):
        self.proxy_registry = registry
        self.proxy_autosync = True

        if eid is None and entity is None:
            raise ValueError(
                "A proxy must be initialized either with an entity ID"
                " or with an entity JSON object containing the ID"
            )

        if entity is not None:
            if eid is None:
                try:
                    eid = entity[self.proxy_schema.id.entity_name]
                except KeyError as e:
                    raise ValueError(
                        "A proxy must be initialized either with an entity ID"
                        " or with an entity JSON object containing the ID"
                    ) from e
            # this is a check which is not really necessary...
            elif self.proxy_schema.id.entity_name in entity:
                if str(eid) != str(entity[self.proxy_schema.id.entity_name]):
                    raise ValueError(
                        "Mismatch between entity ID provided directly and indirectly"
                    )

        if not isinstance(eid, UUID):
            self.proxy_id = UUID(eid)
        else:
            self.proxy_id = eid

        self.proxy_attr = None
        self.proxy_changed = None

    def __init_subclass__(cls, entity=True):
        from .schema import Schema

        if entity:
            cls.proxy_schema = Schema(cls)
        else:
            # cls is not an entity class, check it
            Schema.check_non_entity(cls)

    @classmethod
    def new(
        cls, regspec: Registry | RegistryCatalog, *, autosync: bool = True, **fields
    ):
        """Return a non-affiliated proxy instance, i.e. a 'creation proxy'.

        This proxy instance's id is UUID(int=0), which is indicative
        of a non-valid id. The purpose of such an object is to
        create a new entity, by calling 'proxy_sync()'.

        While the UUID is 0, this entity is not registered in
        the registry. Therefore, multiple such entries can exist.
        However, once the entity is proxy_sync'd, its id changes
        to a proper one and the object is registered in the
        registry.

        Args:
            cls (Type[ProxyClass]): the proxy type for the new proxy.

            regspec (Registry|RegistryCatalog): Registry spec for the new object.
               If regspec is a Registry, its catalog is used to locate a suitable registry
               for the 'cls' proxy type, via regspec.catalog.registry_for(cls).
               If regspec is a catalog, it provides the registry via regspec.registry_for(cls).

            autosync (bool, default: True): When True, the new entity is created (by calling
                proxy_sync() on the proxy) before returning. When False, the new entity
                is not yet created, a call to proxy_sync() must be done later.

                Note that the proxy_aytosync field of the returned proxy is set to true, regardless
                of the value of this parameter.

            fields (dict[str][Any]):  values to initialize the new entity from

        Returns:
            A proxy to a new entity. If autosync if False, the new proxy does not yet correspond
            to an entity.
        """
        from .registry import Registry, RegistryCatalog

        if not hasattr(cls, "proxy_schema"):
            raise TypeError(f"Class {cls.__name__} is not an entity class")

        if isinstance(regspec, RegistryCatalog):
            registry = regspec.registry_for(cls)
        elif isinstance(regspec, Registry):
            registry = regspec.catalog.registry_for(cls)
        else:
            raise TypeError("Expected Registry or RegistryCatalog for regspec")

        schema = cls.proxy_schema
        proxy = cls(registry, eid=UUID(int=0))

        # Validate the given fields
        validated_fields = {
            name: schema.properties[name].validate(proxy, value)
            for name, value in fields.items()
            if name in schema.all_fields
        }

        # For the missing fields, add the default, or ...
        # Adding ..., implies somehow that the field has been deleted (!)
        for name, prop in schema.properties.items():
            if not (prop.isId or prop.isExtras or name in validated_fields):
                defval = prop.missing(proxy=proxy)
                if defval is ...:
                    validated_fields[name] = ...
                else:
                    try:
                        validated_fields[name] = prop.validate(proxy, defval)
                    except ValueError as e:
                        raise ConversionError(prop, "validate") from e
                # if False:  # prop.create_default is not None:
                #    validated_fields[name] = prop.create_default
                # else:
                #    validated_fields[name] = ...

        # Finally, if we have extras, use extras field for any unrecognized
        # items in fields
        if schema.extras is not None:
            validated_fields[schema.extras.name] = {
                ename: schema.extras.item_validator.validate(evalue)
                for ename, evalue in fields.items()
                if ename not in validated_fields
            }

        # Set up the dictionaries of the proxy
        proxy.proxy_attr = validated_fields
        proxy.proxy_changed = dict()

        if autosync:
            proxy.proxy_sync()
        return proxy

    @classmethod
    def new_entity(cls, catalog: RegistryCatalog = None, /, **fields) -> Entity:
        """Return a set of fields for creating a new entity.

        Args
        ----
            fields: dict[str,Any]
        """
        if not hasattr(cls, "proxy_schema"):
            raise TypeError(f"Class {cls.__name__} is not an entity class")
        entity_fields = {}
        for property in cls.proxy_schema.properties.values():
            property.convert_to_create(
                cls,
                fields,
                entity_fields,
                catalog=catalog,
                registry=catalog.registry_for(cls) if catalog else None,
            )
        return entity_fields

    def delete(self, purge: bool = False):
        """Delete the entity and mark the proxy as invalid.
        Entity classes can overload this method, to perform the
        actual API delete. When successful, they can then
        invoke Proxy.delete() to mark this proxy as invalid.
        """
        if self.proxy_state is ProxyState.ERROR:
            return  # Not an error to call delete on purged entity
        if purge:
            self.proxy_is_purged()
        else:
            self.proxy_invalidate(force=True)

    def update(self, **updates: Any):
        """Update a bunch of attributes in a single operation."""
        with deferred_sync(self):
            for name, value in updates.items():
                if value is ...:
                    delattr(self, name)
                else:
                    setattr(self, name, value)

    @property
    def proxy_state(self) -> ProxyState:
        """Return the proxy state"""
        if self.proxy_id is None:
            return ProxyState.ERROR
        elif self.proxy_attr is None:
            return ProxyState.EMPTY
        elif self.proxy_changed is None:
            return ProxyState.CLEAN
        else:
            return ProxyState.DIRTY

    def proxy_invalidate(self, *, force=False):
        """Make this proxy object EMPTY, discarding any entity data.

        If the proxy object is DIRTY, a `InvalidationError` exception
        is raised, unless `force` is true.

        Args:
            force (bool): Invalidate even if DIRTY, without raising exception.
                Deafults to False.

        Raises:
            InvalidationError: If called on a DIRTY proxy with `force` being False
        """
        if self.proxy_id is None:
            raise ErrorState()
        if self.proxy_changed is not None and not force:
            raise InvalidationError()
        self.proxy_attr = self.proxy_changed = None

    def proxy_reset(self):
        """If proxy is EMPTY, do nothing.
        If the proxy is DIRTY, make it CLEAN by restoring the
        values changed since the last sync.
        """
        if self.proxy_id is None:
            raise ErrorState()
        if self.proxy_changed is not None:
            for name, value in self.proxy_changed.items():
                self.proxy_attr[name] = value
            self.proxy_changed = None

    def proxy_from_entity(self, entity: Any):
        """Update the proxy_attr dictionary from a given entity."""
        if self.proxy_id is None:
            raise ErrorState()
        if self.proxy_attr is None:
            self.proxy_attr = dict()
        for prop in self.proxy_schema.properties.values():
            if not prop.isId:
                prop.convert_entity_to_proxy(self, entity)

    def proxy_to_entity(
        self, attrset: set[str] | dict[str, Any] | None = None
    ) -> Entity:
        """Return an entity from the proxy values.

        Note that the entity returned will not contain the id attribute.

        Args:
            attrset (set of property names, optional): If not None,
                determines the set of names to add to the entity.

                Any type of object, where the expression
                `name in attrset` is valid, can be used.

                Use this to only add names of changed properties to
                an entity:
                self.proxy_to_entity(self.proxy_changed)

        Returns:
            entity (dict): An entity dict containing all values
                specified.
        """
        if self.proxy_id is None:
            raise ErrorState()
        entity = dict()
        for prop in self.proxy_schema.properties.values():
            if prop.isId or (attrset is not None and prop.name not in attrset):
                continue
            try:
                prop.convert_proxy_to_entity(self, entity)
            except Exception as e:
                raise ConversionError(prop, "convert_proxy_to_entity") from e
        return entity

    def proxy_is_purged(self):
        """Called to designate that this proxy is referring to a non-existent
        entity and should be marked as such.

        This type of marking happens when an entity is purged.
        """
        if self.proxy_state is ProxyState.ERROR:
            return
        self.proxy_purged_id = self.proxy_id
        self.proxy_registry.purge_proxy(self)
        self.proxy_attr = self.proxy_changed = None

    def proxy_autocommit(self):
        """Try a call to proxy_sync() if the proxy is in auto sync.

        This call is invoked whenever the proxy is changed. If the
        proxy is in autosync mode and DIRTY, the method will try to sync
        the proxy with the API. If this fails, the proxy is reset to CLEAN.

        If the proxy is not DIRTY, or not in autosync mode, the method does nothing.
        """
        if self.proxy_state is ProxyState.DIRTY and self.proxy_autosync:
            try:
                self.proxy_sync()
            except Exception:
                self.proxy_reset()
                raise

    def proxy_sync(self, entity=None):
        """Sync the data between the proxy and the API entity.

        After a sync, the proxy is CLEAN and consistent with the
        underlying entity in the Data Catalog.  This method must be
        overloaded in subclasses, to cater to the details of different
        types of entities.

        In order to sync, the method works as follows:

        1. If the proxy is DIRTY:
            - updates are sent to the API.
            - The API optionally returns a new entity object. If so, override
              the `entity` parameter.
            - Make object CLEAN (by setting proxy_changed to None)

        2. If `entity` is None: load `entity` from API

        3. Update the proxy data from `entity`. This may involve
           updating additional proxies with data contained in the given
           entity.

        For typical operations, the implementation can use the following mathods:
        ```
        self.proxy_from_entity(entity)
        self.proxy_to_entity(attrset) -> entity
        ```
        """
        raise NotImplementedError(self.__class__.__name__ + ".proxy_sync")

    def __pos__(self):
        """Return the proxy itself after performing proxy_sync."""
        if self.proxy_state is ProxyState.CLEAN:
            self.proxy_invalidate()
        return self

    def __repr__(self) -> str:
        typename = type(self).__name__
        state = self.proxy_state.name
        if self.proxy_state is ProxyState.ERROR:
            nid = f"deleted ({getattr(self, 'proxy_purged_id', '**unknown**')})"
        elif self.proxy_state is ProxyState.EMPTY:
            nid = str(self.proxy_id)
        elif self.proxy_schema.name_id is not None:
            nid = self.name
        else:
            nid = str(self.proxy_id)
        return f"<{typename} {nid} {state}>"

    def proxy_to_Series(
        self,
        *,
        sync_empty: bool = True,
        include_null: bool = False,
        include_extras: bool = False,
        simplify: bool = True,
    ) -> Series:
        """Return a pandas Series for this entity.

        The pandas Series index will contain entity fields. The values will be simplified,
        to reflect

        Arguments:
            sync_empty (bool, default=True): call proxy_sync() if state is EMPTY
            include_null (bool, default=False): include fields that evaluate to False
            include_extras (bool, default=False): also include any extras fields
            simplify (bool, default=True): return a more printable, simpler representation
        """
        import pandas as pd

        name = f"{type(self).__name__} ({self.proxy_state.name})"
        if self.proxy_state is ProxyState.ERROR or (
            not sync_empty and self.proxy_state is ProxyState.EMPTY
        ):
            return pd.Series(name=name)

        schema = self.proxy_schema

        def report_field(name):
            return include_null or bool(getattr(self, name, False))

        all_fields = [schema.id.name, *schema.properties]
        if include_extras and schema.extras is not None:
            extras = schema.extras.get(self)
            if extras is not ...:
                all_fields.remove(schema.extras.name)
                all_fields.extend(extras.keys())

        index = [name for name in all_fields if report_field(name)]
        index.sort()

        def simplified(val):
            match val:
                case Proxy(proxy_schema=schema) if schema.name_id is not None:
                    return val.name
                case Proxy():
                    return val.proxy_id
                case _:
                    return val

        def propvalue(name):
            value = getattr(self, name, ...)
            if value is not ... and simplify:
                value = simplified(value)
            return value

        data = [propvalue(name) for name in index]
        return pd.Series(index=index, data=data, name=name, dtype="object")

    @property
    def s(self) -> Series:
        return self.proxy_to_Series()

    @property
    def ss(self) -> Series:
        return self.proxy_to_Series(sync_empty=False)

    @property
    def sl(self) -> Series:
        return self.proxy_to_Series(include_null=True)

    @property
    def sx(self) -> Series:
        return self.proxy_to_Series(include_extras=True)

    @property
    def sxl(self) -> Series:
        return self.proxy_to_Series(include_null=True, include_extras=True)

    @property
    def sraw(self) -> Series:
        return self.proxy_to_Series(
            include_null=True, include_extras=True, simplify=False
        )


@contextmanager
def deferred_sync(*proxies):
    if not all(isinstance(p, Proxy) for p in proxies):
        raise TypeError("All arguments must be entity proxies")
    for i, p in enumerate(proxies):
        if any(q is p for q in proxies[i + 1 :]):
            raise ValueError(
                f"The {i}-th argument appears again; proxies must be entered once"
            )
    saved_autosync = [p.proxy_autosync for p in proxies]
    for p in proxies:
        p.proxy_autosync = False
    try:
        yield proxies
    except Exception:
        for p in proxies:
            if p.proxy_state is not ProxyState.ERROR:
                p.proxy_reset()
        raise
    finally:
        for p, a in zip(proxies, saved_autosync):
            p.proxy_autosync = a

    # This belongs outside the finally clause.
    # It will not be executed if there is an error
    exc = []
    for p in proxies:
        if p.proxy_state is ProxyState.DIRTY:
            try:
                p.proxy_sync()
            except Exception as e:
                p.proxy_reset()
                exc.append((p, e.with_traceback(None)))
        elif p.proxy_state is not ProxyState.ERROR:
            p.proxy_sync()
    if exc:
        raise RuntimeError("Failed to sync, updates reset", exc)



from uuid import uuid4

import pytest
from proxy_utils import ProxyTestObj, TPCatalog

from stelar.client.proxy.tag import TaggableProxy, TagList, TagListField
from stelar.client.vocab import tag_split


@pytest.mark.parametrize(
    "tagspec, tagpair",
    [
        ("foo:bar", ("foo", "bar")),
        ("gg:aa:bar", ("gg:aa", "bar")),
        ("", None),
        ("a", None),
        ("a ", (None, "a ")),
        ("--", (None, "--")),
        ("s" * 101, None),
        (":::::::__", ("::::::", "__")),
    ],
)
def test_tag_split(tagspec, tagpair):
    if tagpair is None:
        with pytest.raises(ValueError):
            tag_split(tagspec)
    else:
        assert tag_split(tagspec) == tagpair


@pytest.mark.parametrize(
    "taglist",
    [
        [],
        {},
        (),
        {"aa", "daltons:joe"},
        ["foo", "bar", "abasw"],
        dict(tag1="aa", tag2=44),
    ],
)
def test_taglist_field(taglist):
    c = TPCatalog()
    did = str(uuid4())
    c.vocabs = [
        {
            "name": "daltons",
            "id": did,
            "tags": [
                {"name": "joe"},
                {"name": "jack"},
                {"name": "averell"},
            ],
        }
    ]

    v = TagListField()
    v.validate(taglist, vocindex=c.vocabulary_index)


@pytest.mark.parametrize("taglist", [["a"], "asdasda", 4, [["aa"]]])
def test_taglist_field_novalidate(taglist):
    v = TagListField()
    with pytest.raises(ValueError):
        v.validate(taglist)


def test_vocabulary_index():
    c = TPCatalog()

    did = str(uuid4())
    c.vocabs = [
        {
            "name": "daltons",
            "id": did,
            "tags": [
                {"name": "joe"},
                {"name": "jack"},
                {"name": "averell"},
            ],
        }
    ]

    assert c.vocabulary_index.dirty
    assert c.vocabulary_index.refresh_count == 0

    c.vocabulary_index.name_to_id["daltons"] == did
    c.vocabulary_index.id_to_name[did] == "daltons"
    assert not c.vocabulary_index.dirty
    assert c.vocabulary_index.refresh_count == 1

    assert c.vocabulary_index.validate_tagspec("daltons:jack")
    assert c.vocabulary_index.validate_tagspec("daltons:joe")
    assert c.vocabulary_index.validate_tagspec("daltons:averell")

    assert c.vocabulary_index.validate_tagspec("fii")
    assert c.vocabulary_index.validate_tagspec("will")
    assert c.vocabulary_index.validate_tagspec("ww")

    assert not c.vocabulary_index.validate_tagspec("doltons:joe")
    assert not c.vocabulary_index.validate_tagspec("daltons:william")


def test_taggable():
    class Foo(TaggableProxy, ProxyTestObj):
        tags = TagList()

    c = TPCatalog()
    c.vocabs = [
        {
            "name": "daltons",
            "id": uuid4(),
            "tags": [
                {"name": "joe"},
                {"name": "jack"},
                {"name": "averell"},
            ],
        }
    ]

    x = Foo.new(c)
    assert x.tags == ()

    x.tags = ("foo", "bar", "daltons:averell")
    assert "foo" in x.tags
    assert "bar" in x.tags
    assert "daltons:averell" in x.tags

    assert Foo.data[x.id]["tags"] == ["foo", "bar", "daltons:averell"]

    # assign by several types of sequences
    x.tags = []
    assert not x.tags

    x.tags = ["foo", "daltons:averell", "bar"]
    assert x.tags == ("foo", "daltons:averell", "bar")

    with pytest.raises(ValueError) as e:
        x.tags = ["doltons:joe"]

    x.proxy_reset()
    assert x.tags == ("foo", "daltons:averell", "bar")



class APIEndpointsV1:

    """
    This class contains a universal definition of the API
    endpoints. This way, any modifications made here will be
    globally updated throughout the client's classes.

    All endpoints contained here will have the api_url of the Client class
    prepended when called through the request(...) method of the BaseAPI parent class.

        e.g.:
        '/auth/token/issue'
        will be configured to be used as
        http://mydomain.eu/stelar/api/v1/auth/token/issue

    Attributes:

    TOKEN_ISSUE (string):   The endpoint for issuing an OAuth2.0 token
    TOKEN_VERIFY (string):  The endpoint for verifying the status of an OAuth2.0 token
    TOKEN_REFRESH (string): The endpoint for issuing an OAuth2.0 token based on a refresh token already available.
    GET_USER (string):      The endpoint for fetching information about a specific user by ID.
    GET_USERS (string):     The endpoint for fetching information about all users in the KLMS.
    GET_DATASET (string):   The endpoint for fetching information about a specific dataset by ID.
    GET_DATASETS (string):  The endpoint for fetching information about all datasets in the KLMS.

    """

    # Used under different HTTP schemes (POST for ISSUE, PUT for REFRESH)

    # Authentication/Authorization Endpoints
    TOKEN_ISSUE = "v1/users/token"  # POST
    TOKEN_REFRESH = "v1/users/token"  # PUT
    TOKEN_VERIFY = "v1/users/token"  # GET
    S3_CREDENTIALS = "v1/users/s3/credentials"  # GET

    GET_USER = "v1/users/"
    GET_USERS = "v1/users"

    # Policy Endpoints
    POST_POLICY = "v1/auth/policy"
    GET_POLICY_REPRESENATION = "v1/auth/policy/representation/"
    GET_POLICY_INFO = "v1/auth/policy/"
    GET_POLICY_LIST = "v1/auth/policy"



from urllib.parse import urlparse

import s3fs
from minio import Minio
from minio.credentials.providers import WebIdentityProvider

from .base import BaseAPI


class S3API(BaseAPI):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Get the MinIO client
        u = urlparse(self._info.s3_api)
        minio_host = u.netloc
        secure = u.scheme == "https"

        def get_jwt():
            if self.token_expired():
                self.reauthenticate()
            return {"access_token": self._token, "expires_in": self._expires_in}

        provider = WebIdentityProvider(
            jwt_provider_func=get_jwt, sts_endpoint=self._info.s3_api
        )
        self._minio_client_provider = provider

        self._minio_client = Minio(
            minio_host, secure=secure, credentials=provider, cert_check=self._tls_verify
        )

    @property
    def s3(self) -> Minio:
        return self._minio_client

    def s3fs(self) -> s3fs.S3FileSystem:
        """Return a file-system handler for the data lake.

        This method returns an instance of the `s3fs.S3FileSystem` class
        that can be used to interact with the data lake.

        Please see <https://github.com/fsspec/s3fs/> for more information.
        """
        acc = self.s3_access_data()
        opts = {
            "endpoint_url": acc["endpoint"],
            "key": acc["key"],
            "secret": acc["secret"],
            "token": acc["token"],
            "use_ssl": acc["secure"],
            "client_kwargs": {"verify": acc["verify"]},
        }
        return s3fs.S3FileSystem(**opts)

    def s3fs_open(self, path: str, mode: str = "rb", **kwargs):
        """Open a file in the data lake.

        This method returns a file-like object that can be used to read or
        write data from the data lake.

        Parameters
        ----------
        path : str
            The path to the file in the data lake.
        mode : str
            The mode in which to open the file. This can be one of the
            following: 'r', 'rb', 'w', 'wb', 'a', 'ab'.
        kwargs : dict
            Additional keyword arguments to pass to the file system.

        Returns
        -------
        file-like
            A file-like object that can be used to read or write data.
        """
        fs = self.s3fs()
        return fs.open(path, mode, **kwargs)

    def s3_access_data(self) -> dict:
        """Return a dict with options for accessing the S3 API."""

        u = urlparse(self.klms_info.s3_api)
        minio_host = u.netloc
        use_ssl = u.scheme == "https"

        acc = self._minio_client_provider.retrieve()
        return {
            "key": acc.access_key,
            "secret": acc.secret_key,
            "token": acc.session_token,
            "endpoint": self.klms_info.s3_api,
            "minio_host": minio_host,
            "secure": use_ssl,
            "verify": self._tls_verify,
            "signature_version": "s3v4",
            "region": "us-east-1",  # This is not needed for MinIO, but if needed this is the default
        }

    def _boto3_client(self):
        """Return a boto3 client for S3 API."""
        import boto3

        acc = self._minio_client_provider.retrieve()
        return boto3.client(
            "s3",
            endpoint_url=self.klms_info.s3_api,
            # region_name="us-east-1",
            aws_access_key_id=acc.access_key,
            aws_secret_access_key=acc.secret_key,
            aws_session_token=acc.session_token,
            config=boto3.session.Config(signature_version="s3v4"),
            verify=False,
        )



import builtins
from uuid import UUID

from .generic import api_call
from .named import NamedProxy
from .proxy import BoolField, DateField, Property, Proxy, ProxyVec, StrField


class MemberList(ProxyVec):
    def __init__(self, client, proxy_type, members, capacities):
        super().__init__(client, proxy_type, members)
        self.capacities = capacities

    def to_df(self, *additional, fields=None, simplify=True):
        df = super().to_df(*additional, fields=fields, simplify=simplify)
        return df.assign(capacity=self.capacities)


class GroupBase(NamedProxy, entity=False):
    """
    Proxy for a STELAR Data Catalog group and organization.
    This is an abstract class. The group subclass is
    defined later.
    """

    is_organization = Property(validator=BoolField)

    created = Property(validator=DateField)
    approval_status = Property(validator=StrField(), updatable=True)

    title = Property(validator=StrField, updatable=True)
    description = Property(validator=StrField, updatable=True)
    image_url = Property(validator=StrField(), updatable=True)

    def get_members(
        self, proxy_type: builtins.type[Proxy], capacity: str | None = None
    ) -> MemberList:  # type: ignore
        """Get the members of the group.

        Args:
            proxy_type: The type of the members.
            capacity: The capacity of the members.
        Returns:
            MemberList, a list of members.
        """

        ac = api_call(self)

        list_members = ac.get_call(self.__class__, "list_members", proxy_type)

        result = list_members(id=str(self.id), capacity=capacity)
        ids = [UUID(entry[0]) for entry in result]
        cap = [entry[2] for entry in result]
        return MemberList(ac.client, proxy_type, ids, cap)

    @property
    def users(self):
        """Get the users of the group."""
        from .user import User

        return self.get_members(User, capacity=None)

    @property
    def datasets(self):
        """Get the datasets of the group."""
        from .dataset import Dataset

        return self.get_members(Dataset, capacity=None)

    @property
    def workflows(self):
        """Get the workflows of the group."""
        from .workflows import Workflow

        return self.get_members(Workflow, capacity=None)

    @property
    def tools(self):
        """Get the tools of the group."""
        from .workflows import Tool

        return self.get_members(Tool, capacity=None)

    @property
    def groups(self):
        """Get the groups of the group."""
        return self.get_members(Group, capacity=None)

    def add(self, member: Proxy, capacity: str = "member"):
        """Add a member to the group.

        Args:
            member: The member to add.
            capacity: The capacity of the member.
        """
        ac = api_call(self)
        add_member = ac.get_call(self.__class__, "add", member.__class__)
        add_member(str(self.id), str(member.proxy_id), capacity=capacity)

    def remove(self, member: Proxy):
        """Remove a member from the group.

        Args:
            member: The member to remove.
        """
        ac = api_call(self)
        member_delete = ac.get_call(self.__class__, "remove", member.__class__)
        member_delete(str(self.id), str(member.proxy_id))


class Group(GroupBase):
    pass


class Organization(GroupBase):
    pass



"""A module containing useful definitions related to STELAR API entities
"""


class MissingParametersError(Exception):
    pass


class DuplicateEntryError(Exception):
    pass


class STELARUnknownError(Exception):
    pass


class EntityNotFoundError(Exception):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Type, TypeVar
from uuid import UUID

import pandas as pd

from .proxy import Proxy

if TYPE_CHECKING:
    from ..client import Client
    from .refs import RefList


ProxyClass = TypeVar("ProxyClass", bound=Proxy)


class ShortenedUUID(UUID):
    def __str__(self):
        return f"{super().__str__()[:5]}..."


def simplified(value, property):
    from .property import Id
    from .refs import Reference

    match property:
        case Id():
            return ShortenedUUID(value.hex)
        case Reference(proxy_type=pt):
            if pt.proxy_schema.name_id:
                return value.name
            elif isinstance(value, Proxy):
                return ShortenedUUID(value.proxy_id.hex)
            elif isinstance(value, ProxyList):
                return f"{pt.proxy_schema.class_name}[{len(value)}]"
            else:
                return value
        case _:
            return value


class ProxyList(Generic[ProxyClass]):
    """
    Base class for "dynamic lists" of proxies.

    The instances maintain lists of IDs and translate list operations
    as if the list actually contained proxy objects.
    """

    def __init__(self, client: Client, proxy_type: Type[ProxyClass]):
        self.client = client
        self.proxy_type = proxy_type
        self.registry = client.registry_for(proxy_type)

    @property
    def coll(self) -> list[UUID]:
        raise NotImplementedError

    def resolve_proxy(self, item):
        """This is the main routine that transforms elements of
        the list to proxies.
        """
        return self.registry.fetch_proxy(item)

    def get_slice(self, slc: slice) -> list[UUID]:
        """Return a list of UUIDs for the given slice.

        This method is used to implement slicing operations on the
        ProxyList. It should be overriden in subclasses depending on
        the underlying implementation the the collection.
        """
        raise ValueError("Slices are not supported yet")

    def __iter__(self):
        for item in self.coll:
            yield self.resolve_proxy(item)

    def __len__(self):
        return len(self.coll)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return ProxyVec(self.client, self.proxy_type, self.get_slice(item))
        else:
            return self.resolve_proxy(self.coll[item])

    def __repr__(self):
        return f"{self.proxy_type.__name__}{repr(self.coll)}"

    def __eq__(self, other):
        try:
            return len(self) == len(other) and (
                all(p == q for p, q in zip(self, other))
            )
        except Exception:
            return False

    @property
    def ids(self):
        return list(self.coll)

    def to_df(self, *additional_fields, fields=None, simplify=True):
        """Generate a pandas dataframe for the list of proxy entities.

        The dataframe is generated by fetching and tabulating a
        subset of fields, for each entity in the list.
        """
        schema = self.proxy_type.proxy_schema

        if fields is None:
            fields = schema.short_list(set(additional_fields))
        elif fields in (True, "all", "ALL"):
            fields = list(schema.all_fields) + list(additional_fields)
        else:
            fields = list(fields) + list(additional_fields)
        data = {field: list() for field in fields}

        for proxy in self:
            for field in fields:
                if simplify:
                    property = schema.all_fields[field]
                    data[field].append(
                        simplified(getattr(proxy, field, pd.NA), property)
                    )
                else:
                    data[field].append(getattr(proxy, field, pd.NA))
        return pd.DataFrame(data=data)

    @property
    def df(self):
        """Return a dataframe over the default fields."""
        return self.to_df()

    @property
    def DF(self):
        """Return a dataframe over all fields."""
        return self.to_df(fields=True, simplify=False)


class ProxyVec(ProxyList):
    """A list of IDs appearing as proxies.

    The underlying data is a list of UUIDs. At each element access, the
    correpsonding element is fetched from the registry.
    """

    def __init__(
        self, client: Client, proxy_type: Type[ProxyClass], members: list[UUID]
    ):
        """Initialize the vector with the client, the proxy type, and the list of IDs.

        Arguments:
            client: the client object
            proxy_type: the type of the proxies
            members: the list of UUIDs
        """
        super().__init__(client, proxy_type)
        self.members: list[UUID] = members

    @property
    def coll(self):
        return self.members

    def get_slice(self, slc: slice) -> list[UUID]:
        return self.members[slc]


class ProxySublist(ProxyList):
    """A proxy class that translates collection operations to operations
    on an entity sub-collection.
    """

    def __init__(self, property: RefList, owner: Proxy):
        super().__init__(owner.proxy_registry.catalog, property.proxy_type)
        self.property = property
        self.owner = owner

    def __delitem__(self, key):
        raise NotImplementedError(
            f"delitem  {self.property.owner.__name__}.{self.property.name}"
        )

    def __iadd__(self, **kwargs):
        raise NotImplementedError(
            f"iadd {self.property.owner.__name__}.{self.property.name}"
        )

    @property
    def coll(self):
        return self.property.get(self.owner)

    def get_slice(self, slc: slice) -> list[UUID]:
        return self.property.get(self.owner)[slc]



from __future__ import annotations

from functools import cached_property
from typing import Iterator

from .generic import GenericCursor, GenericProxy, api_call
from .proxy import BoolField, DateField, Id, Property, StrField, derived_property


class User(GenericProxy):
    id = Id()

    username = Property(validator=StrField())

    fullname = Property(validator=StrField())
    first_name = Property(validator=StrField(), updatable=True)
    last_name = Property(validator=StrField(), updatable=True)

    email = Property(validator=StrField(), updatable=True)
    email_verified = Property(validator=BoolField(), updatable=True)

    @derived_property
    def roles(self, entity):
        return tuple(entity.get("roles", []))

    joined_date = Property(validator=DateField())
    active = Property(validator=BoolField(), updatable=True)

    def add_role(self, role: str):
        """
        Add a role to the user.

        Args:
            role (str): The role to add.
        """
        ac = api_call(self)
        entity = ac.user_add_role(self.proxy_id, role)
        self.proxy_sync(entity)

    def remove_role(self, role: str):
        """
        Remove a role from the user.

        Args:
            role (str): The role to remove.
        """
        ac = api_call(self)
        entity = ac.user_remove_role(self.proxy_id, role)
        self.proxy_sync(entity)

    def append_roles(self, roles: list[str]):
        """
        Append multiple roles to the user.

        Args:
            roles (list[str]): The roles to append.
        """
        ac = api_call(self)
        entity = ac.user_append_roles(self.proxy_id, roles)
        self.proxy_sync(entity)

    def set_roles(self, roles: list[str]):
        """
        Set the user's roles to a new list.

        Args:
            roles (list[str]): The new list of roles.
        """
        ac = api_call(self)
        entity = ac.user_set_roles(self.proxy_id, roles)
        self.proxy_sync(entity)


class UserCursor(GenericCursor[User]):
    def __init__(self, client):
        super().__init__(client, User)

    def create(self, **kwargs) -> User:
        """
        Create a new user.

        Args:
            **kwargs: The user attributes to set.

        Returns:
            User: The created user.
        """
        # The user_create method is needed to support providing
        # the password as a keyword argument.
        ac = api_call(self.client)
        result = ac.user_create(**kwargs)
        return self.fetch_proxy_for_entity(result)

    @cached_property
    def current_user(self) -> User:
        """
        The current user of the client.

        Returns:
            dict: A dictionary containing the user's information.
        """
        return self.get(self.client._username)

    def fetch_list(self, *, limit: int, offset: int) -> list[str]:
        ac = api_call(self.client)
        result = ac.user_list(limit=limit, offset=offset)
        return result

    def fetch(self, *, limit: int, offset: int) -> Iterator[User]:
        registry = self.client.registry_for(User)
        ac = api_call(self.client)
        result = ac.user_fetch(limit=limit, offset=offset)

        for entity in result:
            yield registry.fetch_proxy_for_entity(entity)

    def roles(self) -> list[str]:
        """
        Fetch the list of roles available in the system.

        Returns:
            list[str]: A list of role names.
        """
        ac = api_call(self.client)
        return [role["name"] for role in ac.roles_fetch()]



from __future__ import annotations

from typing import Any

from .exceptions import EntityError
from .property import Property
from .proxy import Proxy


class derived_property(Property):
    """Derived properties are used to cache some 'expensive' value into the proxy.

    The value is recomputed at each proxy_sync(), during a call to 'proxy_from_entity()'.
    This value is then returned by the getter. Additional 'vanilla python properties' or
    methods can be declared in the usual way, to implement richer functionality.
    """

    def __init__(self, method=None, **kwargs):
        super().__init__(updatable=False, optional=False, create_default=None, **kwargs)
        self.method = method

    # Disable mutating ops
    def set(self, obj, value):
        pass

    def __call__(self, method):
        self.method = method
        return self

    def convert_entity_to_proxy(self, proxy: Proxy, entity: Any, **kwargs):
        proxy.proxy_attr[self.name] = self.method(proxy, entity)

    def convert_proxy_to_entity(self, proxy: Proxy, entity: dict, **kwargs):
        pass

    def convert_to_create(
        self, proxy_type, create_props: dict, entity_props: dict, **kwargs
    ):
        if self.name in create_props:
            raise EntityError("Cannot provide derived property for creation")



def test_fetch_users(testcli):
    # Fetch all users via the API
    raw_users = testcli.GET("v1/users").json()["result"]

    # Fetch all users via the client
    users = testcli.users[::]

    # Check that the number of users is the same
    assert len(users) == len(raw_users)

    raw_user_names = {u["username"] for u in raw_users}
    user_names = {u.username for u in users}

    assert raw_user_names == user_names


def test_update_dummy_user_name(testcli):
    c = testcli
    du = c.users["dummy_user2"]

    assert du.first_name == "Foo"
    assert du.last_name == "Manchu"

    du.first_name = "Bar"
    assert du.first_name == "Bar"
    assert du.last_name == "Manchu"
    assert du.fullname == "Bar Manchu"

    du.first_name = "Foo"
    assert du.fullname == "Foo Manchu"


def test_update_dummy_user_email(testcli):
    c = testcli
    du = c.users["dummy_user2"]

    assert du.email == "dumb2@dumbville.com"
    assert du.email_verified is True

    du.email = "dumb_2@dumbville.com"
    assert du.email_verified is True

    du.email = "dumb2@dumbville.com"
    assert du.email_verified is True

    du.email_verified = False
    assert du.email_verified is False
    du.email_verified = True


def test_read_only_username(testcli):
    c = testcli
    du = c.users["dummy_user2"]

    assert du.username == "dummy_user2"


def test_read_roles(testcli):
    c = testcli
    roles = c.users.roles()

    assert all(r in roles for r in ["admin", "pushers", "pullers"])


def test_read_user_roles(testcli):
    c = testcli
    du = c.users["dummy_user2"]

    du.set_roles([])
    assert du.roles == ()
    assert len(du.roles) == 0

    du.add_role("pushers")

    assert du.roles == ("pushers",)
    assert len(du.roles) == 1

    du.set_roles(["pullers", "pushers"])

    assert set(du.roles) == set(("pullers", "pushers"))
    assert len(du.roles) == 2

    du.remove_role("pushers")
    assert du.roles == ("pullers",)
    assert len(du.roles) == 1



from .base import BaseAPI
from .package import PackageCursor
from .process import ProcessCursor
from .tasks import TaskCursor
from .tool import ImageRegistryTokenCursor, ToolCursor
from .workflows import Workflow


class WorkflowsAPI(BaseAPI):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.processes = ProcessCursor(self)
        self.tasks = TaskCursor(self)
        self.workflows = PackageCursor(self, Workflow)
        self.tools = ToolCursor(self)
        self.image_registry_tokens = ImageRegistryTokenCursor(self)



import pytest

from stelar.client import ErrorState, ProxyState


def test_dataset_create(testcli):
    c = testcli

    d = c.datasets.create(name="test_dataset_create", title="A test")
    assert d.title == "A test"

    d.proxy_invalidate()
    assert d.title == "A test"

    d.delete(purge=True)
    assert d.proxy_state is ProxyState.ERROR


def test_dataset_cursor_not_found(testcli):
    c = testcli

    with pytest.raises(KeyError):
        d = c.datasets["test_not_found"]

    d = c.datasets.create(name="test_not_found")
    assert c.datasets["test_not_found"] is d
    assert c.datasets[d.id] is d


def test_dataset_add_resources(testcli):
    c = testcli

    d = c.datasets.create(name="test_dataset_add_resources")
    r1 = c.resources.create(dataset=d, name="my name")
    r2 = c.resources.create(dataset=d, name="my name")

    assert r1.dataset is d
    assert r1.name == "my name"

    assert r2.dataset is d
    assert r2.name == "my name"


def test_dataset_remove_resources(testcli):
    c = testcli

    d = c.datasets.create(name="test_dataset_add_resources")
    r1 = c.resources.create(dataset=d, name="my name")
    r2 = c.resources.create(dataset=d, name="my name")

    assert r1.dataset is d
    assert r1.name == "my name"

    assert r2.dataset is d
    assert r2.name == "my name"

    assert d.resources == [r1, r2]
    r1.delete()
    assert d.resources == [r2]
    r2.delete()
    assert d.resources == []


def test_dataset_delete(testcli):
    c = testcli

    d = c.datasets.create(name="test_dataset_add_resources")
    r1 = c.resources.create(dataset=d, name="my name")
    r2 = c.resources.create(dataset=d, name="my name")

    assert r1.dataset is d
    assert r1.name == "my name"

    assert r2.dataset is d
    assert r2.name == "my name"

    d.delete()
    assert d.state == "deleted"

    assert r1.dataset is d
    assert r2.dataset is d


def test_dataset_purge(testcli):
    c = testcli

    d = c.datasets.create(name="test_dataset_add_resources")
    r1 = c.resources.create(dataset=d, name="my name")
    r2 = c.resources.create(dataset=d, name="my name")

    assert r1.dataset is d
    assert r1.name == "my name"

    assert r2.dataset is d
    assert r2.name == "my name"

    d.delete(purge=True)
    assert d.proxy_state is ProxyState.ERROR
    with pytest.raises(ErrorState):
        d.proxy_sync()
    with pytest.raises(ErrorState):
        d.title
    with pytest.raises(ErrorState):
        d.proxy_reset()
    with pytest.raises(ErrorState):
        d.proxy_invalidate()

    assert r1.proxy_state is ProxyState.ERROR
    assert r2.proxy_state is ProxyState.ERROR


def test_dataset_add_resource(testcli):
    c = testcli

    d = c.datasets.create(name="test_dataset_add_resources")
    r1 = d.add_resource(name="my name")

    # With more attributes and an extra field
    r2 = d.add_resource(name="my name also", url="http://foo", foo="bar")

    assert r1.dataset is d
    assert r1.name == "my name"

    assert r2.dataset is d
    assert r2.name == "my name also"
    assert r2.url == "http://foo"
    assert r2.foo == "bar"


def test_dataset_extras(testcli):
    c = testcli

    d = c.datasets.create(name="test_dataset_add_resources")
    d.add_resource(name="my name")

    assert d.extras == {}

    d.msg1 = "This is a message"
    assert d.extras == {"msg1": "This is a message"}

    d.proxy_invalidate()

    assert d.msg1 == "This is a message"
    del d.msg1
    assert not hasattr(d, "msg1")


def test_resource_extras(testcli):
    c = testcli
    d = c.datasets.create(name="test_dataset_add_resources")
    r = d.add_resource(name="my name")

    jcs = [
        {"a": 10, "b": ["hello", "world"]},
        "foobar",
        "",
        3.14,
        ["hello", {"type": "good"}, "world"],
    ]

    for v in jcs:
        r.foo = v
        r.proxy_invalidate
        assert r.foo == v



"""Basic support for tasks in the client."""

from __future__ import annotations

from typing import TYPE_CHECKING, Generator
from uuid import UUID

from .dataset import Dataset
from .group import Organization
from .resource import Resource
from .utils import convert_proxy_id_to_str

if TYPE_CHECKING:
    # Type definitions for
    from .tool import Tool

    ToolSpec = Tool | UUID | str | None
    ImageSpec = str | None
    ExistingDataset = Dataset | UUID | str


class TaskSpec:
    """A task spec is used to create tasks.

    Contrary to other entities, Task creation is more complicated than simply
    providing values for some scalar attributes. Tasks, once created are
    not really editable; therefore, all task fields must be specified
    at task creation time.

    This class helps by collecting arguments needed to initialize a new Task.
    """

    def __init__(
        self, tool: ToolSpec = None, *, image: ImageSpec = None, name: str = None
    ):
        from .tool import Tool

        self.tool = convert_proxy_id_to_str(tool, Tool)
        if isinstance(image, str | None):
            self.image = image
        else:
            raise TypeError("Expected a string as image")

        if not isinstance(name, str | None):
            raise TypeError("Expected a string as name")

        self.name = name
        self.datasets = {}
        self.inputs = {}
        self.outputs = {}
        self.parameters = {}

    @classmethod
    def from_json(cls, json_spec: dict) -> TaskSpec:
        """Create a TaskSpec from a JSON specification.

        Args:
            json_spec (dict): the JSON specification for the task spec.
        """
        tool = json_spec.get("tool")
        image = json_spec.get("image")
        name = json_spec.get("name", "task")

        ts = cls(tool=tool, image=image, name=name)
        ts.datasets = json_spec.get("datasets", {})
        ts.inputs = json_spec.get("inputs", {})
        ts.outputs = json_spec.get("outputs", {})
        ts.parameters = json_spec.get("parameters", {})

        return ts

    def is_external(self) -> bool:
        """Check if this task spec is local (i.e., not using a remote tool)."""
        return self.tool is None and self.image is None

    def d(
        self, alias: str, dset: Dataset | UUID | str | None = None, **dspec
    ) -> TaskSpec:
        """Add a dataset alias in the task spec.

        An alias stands for either an existing dataset or for the
        spec of a future dataset.

        Dataset aliases can be useful in specifying input and output arguments
        to a task.

        Args:
            alias (str): the dataset alias name
            dset  (Dataset|UUID|str|None): specify an existing dataset for this alias
            dspec (dict): the spec for a future dictionary.
        """

        if dset is not None:
            if dspec:
                raise ValueError("Cannot provide both a dataset and a new dataset spec")
            self.datasets[alias] = convert_proxy_id_to_str(dset, Dataset)
        else:
            if "name" not in dspec:
                raise ValueError("A dataset spec must provide a name")
            if "owner_org" not in dspec:
                dspec["owner_org"] = "stelar-klms"
            else:
                dspec["owner_org"] = convert_proxy_id_to_str(
                    dspec["owner_org"], Organization, nullable=False
                )
            self.datasets[alias] = dspec

        return self

    def process_inspec_entry(
        self, inspec: str | UUID | Resource | list[str | UUID | Resource]
    ) -> Generator[str]:
        if isinstance(inspec, (Resource, UUID, str)):
            yield convert_proxy_id_to_str(inspec, Resource)
        elif isinstance(inspec, list | tuple):
            for i in inspec:
                yield from self.process_inspec_entry(i)
        else:
            raise TypeError("Unknown type for input spec")

    def process_inspec(self, inspec) -> list[str]:
        return list(self.process_inspec_entry(inspec))

    def i(self, **input_specs):
        for name, inspec in input_specs.items():
            self.inputs[name] = self.process_inspec(inspec)
        return self

    def process_outspec(self, outspec: str | dict):
        # Output is either an S3 URI (str)
        # or a dict. No processing here (yet!)
        if isinstance(outspec, str):
            return {"url": outspec}
        elif isinstance(outspec, dict):
            match outspec:
                case {
                    "url": url,
                    "resource": dict(),
                    "dataset": str(),
                }:
                    return outspec
                case {"url": url, "resource": res}:
                    return {
                        "url": url,
                        "resource": convert_proxy_id_to_str(res, Resource),
                    }
                case {"url": url}:
                    return outspec
                case _:
                    raise ValueError("Bad format for output spec", outspec)
        else:
            raise TypeError(
                f"Bad type {type(outspec)} for output spec, expected string or dict"
            )

    def o(self, **ospec) -> TaskSpec:
        for name, osp in ospec.items():
            self.outputs[name] = self.process_outspec(osp)
        return self

    def p(self, **params) -> TaskSpec:
        self.parameters |= params
        return self

    def spec(self) -> dict:
        """Return the task spec as a dictionary.

        Returns:
            dict: the task spec as a dictionary.
        """
        s = {}

        if self.tool is not None:
            s["tool"] = self.tool

        if self.image is not None:
            s["image"] = self.image

        s["datasets"] = self.datasets.copy()
        s["inputs"] = self.inputs.copy()
        s["outputs"] = self.outputs.copy()
        s["parameters"] = self.parameters.copy()
        s["name"] = self.name if self.name is not None else "unnamed"
        return s



from uuid import UUID, uuid4

from stelar.client.proxy import Proxy, ProxySynclist, Registry, RegistryCatalog


class TPRegistry(Registry):
    def __init__(self, catalog, proxy_type):
        super().__init__(catalog, proxy_type)

    def fetch(self, eid=None):
        if eid is None:
            eid = uuid4()
        return self.proxy_type(self, eid)


class TPCatalog(RegistryCatalog):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.vocabs = []

    def fetch_active_vocabularies(self):
        return self.vocabs

    def registry_for(self, cls):
        if cls not in self.registry_catalog:
            self.add_registry_for(cls, TPRegistry(self, cls))
        return super().registry_for(cls)


##################################################
#  An 'abstract' subclass of ProxyObj which does
#  not implement an entity.
##################################################
class ProxyTestObj(Proxy, entity=False):
    data = {}
    patch = {}

    def proxy_sync(self, entity=None):
        myid = self.proxy_id

        if self.proxy_id == UUID(int=0):
            # Populate sync list
            psl = ProxySynclist()
            psl.on_create_proxy(self)

            # Create the entity
            schema = self.proxy_schema
            entity = self.proxy_to_entity()

            # ------------------------------
            # Simulate entity creation

            # 1. Generate an ID
            myid = uuid4()
            entity[schema.id.entity_name] = str(myid)

            # 2. Fill up non-provided values with defaults
            for prop in schema.properties.values():
                if prop.entity_name not in entity:
                    v = prop.validator
                    entity[prop.entity_name] = v.convert_to_entity(
                        v.default_value(),
                        vocindex=self.proxy_registry.catalog.vocabulary_index,
                    )

            self.data[myid] = entity
            # -------------------------------

            # We add this proxy to its registry!
            # This will call proxy_sync recursively !
            self.proxy_registry.register_proxy_for_entity(self, entity)
            self.proxy_from_entity(entity)
            self.proxy_changed = None
            # Sync around
            psl.sync()
            return

        if self.proxy_changed is not None:
            entity = self.proxy_to_entity()
            self.data[myid] = entity
            self.patch[myid] = self.proxy_to_entity(self.proxy_changed)
            self.proxy_changed = None

        if entity is None:
            entity = self.data[myid]

        self.proxy_from_entity(entity)



import pytest
from proxy_utils import ProxyTestObj, TPCatalog

from stelar.client import deferred_sync
from stelar.client.proxy import IntField, Property, Proxy, ProxyState
from stelar.client.proxy.derived import derived_property
from stelar.client.proxy.property import DictProperty


def test_derived_property():
    class Foo(ProxyTestObj):
        a = Property(validator=IntField, updatable=True)

        @derived_property
        def two_a(self, entity):
            return 2 * entity["a"]

    c = TPCatalog()

    x = Foo.new(c, a=5)
    assert x.a == 5
    assert x.two_a == 10

    x.a = 20
    assert x.two_a == 40

    with deferred_sync(x):
        x.a = 1
        assert x.two_a == 40
    assert x.two_a == 2


def test_inherited_property():
    class Foo(ProxyTestObj, entity=False):
        a = Property(validator=IntField, updatable=True)

    class Bar(Foo):
        b = Property(validator=IntField, updatable=True)

    assert hasattr(Bar, "proxy_schema")
    assert not hasattr(Foo, "proxy_schema")

    assert set(Bar.proxy_schema.properties) == {"a", "b"}
    assert Bar.proxy_schema.abstract_base_classes == [Foo, ProxyTestObj, Proxy]

    c = TPCatalog()

    with pytest.raises(TypeError):
        Foo.new(c, a=1)

    x = Bar.new(c, a=1, b=3)
    assert x.a == 1
    assert x.b == 3


def test_dict_property():
    class Foo(ProxyTestObj):
        a = DictProperty(str, str, updatable=True)
        b = DictProperty(str, int, updatable=True)
        c = DictProperty(int, str, updatable=False)
        d = DictProperty(int, int, nullable=True, updatable=True)

    c = TPCatalog()

    x = Foo.new(c, a={}, b={}, c={5: "low"}, d=None)

    assert x.a == {}
    x.a["high"] = "low"
    assert x.proxy_state is ProxyState.CLEAN
    assert x.a["high"] == "low"

    assert x.b == {}
    x.b["high"] = 1
    assert x.proxy_state is ProxyState.CLEAN
    assert x.b["high"] == 1

    with pytest.raises(ValueError):
        x.a["high"] = 1
    assert x.proxy_state is ProxyState.CLEAN
    assert x.a["high"] == "low"
    assert x.b["high"] == 1

    with pytest.raises(ValueError):
        x.b["high"] = "low"
    assert x.proxy_state is ProxyState.CLEAN
    assert x.a["high"] == "low"
    assert x.b["high"] == 1

    with pytest.raises(ValueError):
        x.a["high"] = None
    assert x.proxy_state is ProxyState.CLEAN

    assert x.c == {5: "low"}
    with pytest.raises(AttributeError):
        x.c[2] = "low"
    with pytest.raises(AttributeError):
        x.c[5] = "high"
    with pytest.raises(AttributeError):
        x.c[2] = None

    assert x.d is None
    x.d = {1: 2}
    assert x.d == {1: 2}
    x.d[4] = 5
    assert x.d == {1: 2, 4: 5}
    x.d = None
    assert x.d is None


def test_dict_property_ops():
    class Foo(ProxyTestObj):
        a = DictProperty(str, str, updatable=True)
        b = DictProperty(str, int, updatable=True)
        c = DictProperty(int, str, updatable=False)
        d = DictProperty(int, int, nullable=True, updatable=True)

    c = TPCatalog()

    x = Foo.new(
        c,
        a={"a" + k: k for k in "abc"},
        b={u: ord(u) for u in "abcdefg"},
        c={n: f"v{n}" for n in range(3)},
        d=None,
    )

    # clear get iter pop update setdefault
    assert x.a.get("a") is None
    assert x.a.get("a", "default") == "default"
    assert x.a.get("aa", 5) == "a"
    assert x.a.get("ac") == "c"
    assert len(x.a) == 3
    assert list(x.a) == ["aa", "ab", "ac"]
    assert list(x.a.keys()) == ["aa", "ab", "ac"]
    assert list(x.a.values()) == ["a", "b", "c"]
    assert list(x.a.items()) == [("aa", "a"), ("ab", "b"), ("ac", "c")]
    assert x.a.pop("ab") == "b"
    assert "ab" not in x.a
    assert x.a.setdefault("ab", "b") == "b"
    assert x.a.setdefault("ab", "c") == "b"
    x.a.update({"ab": "b", "ac": "c"})
    assert x.a == {"aa": "a", "ab": "b", "ac": "c"}
    x.a.clear()
    assert x.a == {}

    assert x.b == {"a": 97, "b": 98, "c": 99, "d": 100, "e": 101, "f": 102, "g": 103}
    assert x.b.get("a") == 97
    assert x.b.get("z") is None
    assert x.b.get("z", 5) == 5
    assert len(x.b) == 7

    assert x.c == {0: "v0", 1: "v1", 2: "v2"}
    for k in range(3):
        assert x.c[k] == f"v{k}"
        assert k in x.c
    assert 3 not in x.c
    assert len(x.c) == 3
    for k, i in enumerate(sorted(x.c)):
        assert k == i



"""
    Setup file for stelar_client.
    Use setup.cfg to configure your project.

    This file was generated with PyScaffold 4.6.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""

from setuptools import setup

if __name__ == "__main__":
    try:
        # setup(version="0.1.10")
        setup(use_scm_version=True, setup_requires=["setuptools_scm"])
    except Exception as e:  # noqa
        import traceback

        traceback.print_exc()

        print(
            "\n\nAn error occurred while building the project, "
            "please ensure you have the most updated version of setuptools, "
            "setuptools_scm and wheel with:\n"
            "   pip install -U setuptools setuptools_scm wheel\n\n"
        )
        raise



from stelar.client import Rel


def test_rel_enum():
    """Test the Rel enum."""

    for rel in Rel:
        assert isinstance(rel, Rel)
        assert isinstance(rel.value, str)
        assert rel.name.lower() == rel.value

    for rel in Rel:
        assert rel.peer().peer() is rel
        assert rel.is_canonical() or rel.peer().is_canonical()


def test_relationship_create(testcli):
    c = testcli

    rels = c.GET("v2/relationships/iris").json()["result"]
    assert len(rels) > 0



import jmespath
import pytest

from stelar.client import Dataset, Process, Tool, Workflow


@pytest.mark.parametrize(
    "proxy_type",
    [Dataset, Tool, Process, Workflow],
)
def test_search_for_type(testcli, proxy_type):
    cursor = testcli.cursor_for(proxy_type)

    # Create a query that counts all entities.
    query = cursor.search()

    ecount = query["count"]
    assert ecount > 0
    assert len(query["results"]) <= ecount
    assert all(e["type"] == proxy_type.__name__.lower() for e in query["results"])

    # Create queries that sort by name ascending and descending, and compare
    names_asc = jmespath.search(
        "results[*].name", cursor.search(sort="name asc", fl=["name"], limit=ecount)
    )
    names_desc = jmespath.search(
        "results[*].name", cursor.search(sort="name desc", fl=["name"], limit=ecount)
    )
    assert names_asc == list(reversed(names_desc))


def test_search_spatial(testcli):
    c = testcli
    try:
        c.datasets["test_dataset"].delete(purge=True)
    except Exception:
        pass

    d = c.datasets.create(name="test_dataset", title="A test")

    # Check that we do not have a spatial field.
    assert d.spatial is None

    # Check that we do not retrieve the dataset in a spatial query.
    fetched_names = jmespath.search(
        "results[*].name", c.datasets.search(bbox=[-180, -90, 180, 90], fl=["name"])
    )
    assert d.name not in fetched_names



from __future__ import annotations

from typing import TYPE_CHECKING

from .fieldvalidation import AnyField, StrField
from .property import Property
from .proxy import Proxy

if TYPE_CHECKING:
    # from ..client import Client
    # from .schema import Schema
    pass


class ExtrasProperty(Property):
    def __init__(self, **kwargs):
        super().__init__(
            updatable=False, optional=False, validator=AnyField(default={}), **kwargs
        )
        self.isExtras = True
        self.item_validator = StrField(nullable=False)

    def autodoc(self, doc, repr_type, repr_constraints):
        return """\
The field holding extras (additional fields) for this entity.
This field should not be accessed directly, instead its contents
are available as normal attributes.
"""

    def get(self, obj):
        """Low-level getter"""
        if obj.proxy_attr is None:
            obj.proxy_sync()
        return obj.proxy_attr[self.name]

    def __get__(self, obj, obj_type=None):
        return self.get(obj).copy()

    def convert_entity_to_proxy(self, proxy: Proxy, entity, **kwargs):
        entity_extras = entity.get(self.entity_name, {})
        proxy_extras = entity_extras.copy()  # Do we need a copy here?
        proxy.proxy_attr[self.name] = proxy_extras

    def convert_proxy_to_entity(self, proxy: Proxy, entity: dict, **kwargs):
        proxy_extras = proxy.proxy_attr[self.name]
        if proxy_extras is ...:
            return
        entity_extras = proxy_extras.copy()  # Do we need a copy here?
        entity[self.entity_name] = entity_extras

    def convert_to_create(self, proxy_type, create_props, entity_props, **kwargs):
        schema = proxy_type.proxy_schema

        # Collect all entries that do not appear in the schema
        entity_extras = {
            key: self.item_validator.validate(value, **kwargs)
            for key, value in create_props.items()
            if key not in schema.all_fields and value is not None
        }
        entity_props[self.entity_name] = entity_extras


class ExtrasProxy(Proxy, entity=False):
    """The class implements a Proxy with an 'extras' field.

    There are two API variants for extra arguments in the
    CKAN data catalog:
    * Resource entities allow additional fields in their objects.::

        "foo": "bar"

    * Packages, Groups and Organizations maintain a separate
      field called extras, whose format is a list of dicts.::

        "extras": [{"key": "foo", "value": "bar"}]

    Our own implementation follows the resource approach, by utilizing
    dynamic attributes.
    """

    # def __init__(self, *args, **kwargs):
    #    for a in [
    #        'proxy_id', 'proxy_attr', 'proxy_changed',
    #        'proxy_purged_id', 'proxy_registry', 'proxy_autosync'
    #    ]:
    #        object.__setattr__(self, a, None)

    def __getattr__(self, attr):
        try:
            return self.proxy_schema.extras.get(self)[attr]
        except KeyError as e:
            raise AttributeError(attr) from e

    def __setattr__(self, attr, value):
        if attr.startswith("proxy_") or attr in self.proxy_schema.all_fields:
            return object.__setattr__(self, attr, value)

        extras_property = self.proxy_schema.extras
        value = extras_property.item_validator.validate(value)
        extras_property.touch(self)
        extras_property.get(self)[attr] = value
        self.proxy_autocommit()

    def __delattr__(self, attr):
        if attr.startswith("proxy_") or attr in self.proxy_schema.all_fields:
            return object.__delattr__(self, attr)

        extras_property = self.proxy_schema.extras
        extras = extras_property.get(self)
        if attr in extras:
            extras_property.touch(self)
            del extras[attr]
            if self.proxy_autosync:
                self.proxy_sync()
        else:
            raise AttributeError(attr)



import pathlib

import pandas as pd

from stelar.client import pdutils
from stelar.client.mutils import is_s3url, s3spec_to_pair
from stelar.client.pdutils import infer_format

from .package import PackageCursor, PackageProxy
from .proxy import Property, StrField
from .reprstyle import dataset_to_html
from .resource import Resource
from .spatial import GeoJSON
from .utils import client_for


class Dataset(PackageProxy):
    """
    A proxy of a STELAR dataset.
    """

    title = Property(validator=StrField, updatable=True)
    url = Property(validator=StrField(nullable=True), updatable=True)
    version = Property(
        validator=StrField(nullable=True, maximum_len=100), updatable=True
    )

    # The spatial property
    spatial = Property(validator=GeoJSON(nullable=True), updatable=True)

    # N.B. This has been removed from the schema
    # license_id = Property(validator=StrField(nullable=True), updatable=True)
    # profile

    # relationships_as_object
    # relationships_as subject

    def add_dataframe(
        self, df: pd.DataFrame, s3path: str, format: str = None, **kwargs
    ) -> Resource:
        """Add a DataFrame as a resource to the dataset.

        Args:
            df (pd.DataFrame): The DataFrame to add.
            s3path (str): The S3 path to save the DataFrame.
            format (str): The format of the file. If not specified, an attempt will
                be done to infer it.
            **kwargs: Additional keyword arguments to pass to the write_dataframe function
        """

        # Infer the format of the file to save
        fmt = infer_format(s3path, format)
        if fmt is None:
            raise ValueError("Could not infer the format of the file to save.")

        # Collect properties of the new resource
        if not is_s3url(s3path):
            raise ValueError("The path must be an S3 URL.")
        bucket, path = s3spec_to_pair(s3path)
        stem = pathlib.PurePosixPath(path).stem

        rsrc = self.add_resource(
            name=stem,
            url=s3path,
            format=fmt,
            mimetype=f"application/{fmt}",
            description=df.describe().to_json(),
            columns=df.columns.tolist(),
            rows=len(df),
            size=df.memory_usage().sum(),
            relation="owned",
        )

        try:
            pdutils.write_dataframe(client_for(self), df, s3path, format=fmt, **kwargs)
        except Exception as e:
            rsrc.delete()
            raise e

        return rsrc

    def read_dataframe(self, format: str | None = None, **kwargs):
        """Read the dataset as a DataFrame.

        Note: the dataframe need not be stored in S3 Storage.

        Args:
            format (str): The format of the file to read. If not specified, the format will be
                inferred from the file extension.
            kwargs (dict): Additional keyword arguments to pass to the read.

        Returns:
            pd.DataFrame: The DataFrame read from the dataset.
        """

        if not self.url:
            raise ValueError("The dataset URL is nil.")

        if (not format) and hasattr(self, "format"):
            format = self.format

        fmt = infer_format(self.url, format)
        if fmt is None:
            raise ValueError("Could not infer the format of the file to save.")

        return pdutils.read_dataframe(client_for(self), self.url, format=fmt, **kwargs)

    def export_zenodo(self) -> dict:
        """Export the dataset to Zenodo.

        This method will create and return a new Zenodo record for the dataset.
        The method returns the Zenodo record as a dictionary.

        Returns:
            dict: The Zenodo record for the dataset.
        """
        ac = client_for(self).api_call()
        return ac.dataset_export_zenodo(self.id)

    def _repr_html_(self):
        return dataset_to_html(self)

    def __disabled_str__(self):
        dataset_info = f"""Title: {self.title} | Dataset ID: {self.id} | Name: {self.name} | Tags: {self.tags} | Modified Date: {self.modified_date}\nDataset Resources:\n"""
        if self.resources:
            for resource in self.resources:
                dataset_info += "\t" + str(resource) + "\n"
        else:
            dataset_info += "\tNo Resources Associated"
        return dataset_info


class DatasetCursor(PackageCursor[Dataset]):
    def __init__(self, client):
        super().__init__(client, Dataset)

    def publish_file(
        self,
        s3file: str,
        format: str | None = None,
        *,
        resource: dict | None = {},
        **dataset_properties,
    ) -> Dataset:
        """Publish a new dataset in the catalog for a single file in Storage.

        The main input to this call is a single file in Storage, identified by its S3 URL.
        By default, this call will create a new dataset in the catalog and a new resource,
        which will be initialized by provided properties and also by infering properties
        by analyzing the given S3 URL.

        If the :code:`resource` parameter is set explicitly to None, no resource will be created;
        instead, the dataset will be assigned the given S3 URL as its URL.

        Args:
            s3path (str): The S3 path to the file.
            format (str): The format of the file. If not specified, an attempt will
                be done to infer it.
            resource_properties (dict): Properties of the new resource. If this is set explicitly
                to None, no resource will be created; instead, the dataset
            **dataset_properties: Properties of the new dataset.
        """

    def publish_dataframe(
        self,
        df: pd.DataFrame,
        s3path: str,
        format: str | None = None,
        *,
        write={},
        **properties,
    ) -> Dataset:
        """Publish a DataFrame as a new dataset.

        The dataframe will be stored at the given path in the format specified
        by the 'format' argument. If the format is not specified, an attempt will
        be done to infer it from the file extension of the s3path.

        Additional arguments to the pandas write method (DataFrame.to_{format})
        can be passed using the 'write' argument.

        Args:
            df (pd.DataFrame): The DataFrame to publish.
            s3path (str): The S3 path to save the DataFrame.
            format (str): The format of the file. If not specified, an attempt will
                be done to infer it.
            write (dict): Keyword arguments to pass to the write_dataframe function
            **properties: Properties of the new dataset.
        """

        # Infer the format of the file to save
        fmt = infer_format(s3path, format)
        if fmt is None:
            raise ValueError("Could not infer the format of the file to save.")

        # Collect properties of the new resource
        if not is_s3url(s3path):
            raise ValueError("The path must be an S3 URL.")
        bucket, path = s3spec_to_pair(s3path)
        stem = pathlib.PurePosixPath(path).stem

        if "name" not in properties:
            properties["name"] = stem
        if "title" not in properties:
            properties["title"] = stem
        if "author" not in properties:
            properties["author"] = self.client.users.current_user.fullname
            properties["author_email"] = self.client.users.current_user.email
        if "maintainer" not in properties:
            properties["maintainer"] = self.client.users.current_user.fullname
            properties["maintainer_email"] = self.client.users.current_user.email

        properties["url"] = s3path
        properties["columns"] = df.columns.tolist()
        properties["rows"] = len(df)
        properties["size"] = df.memory_usage().sum()
        properties["format"] = fmt
        properties["mimetype"] = f"application/{fmt}"
        properties["description"] = df.describe().to_json()

        dataset = self.create(**properties)
        try:
            pdutils.write_dataframe(self.client, df, s3path, format=fmt, **write)
        except Exception as e:
            dataset.delete(purge=True)
            raise e
        return dataset
